"""Functionality common to displaying dataset and variables metadata."""

from __future__ import annotations

import logging
import urllib
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import cast

import arrow
import ssb_dash_components as ssb
from dapla_metadata.datasets import enums
from dapla_metadata.datasets import model
from dash import dcc
from dash import html

from datadoc_editor import state
from datadoc_editor.constants import DELETE_SELECTED
from datadoc_editor.constants import DROPDOWN_DELETE_OPTION
from datadoc_editor.frontend.components.identifiers import ADD_USE_RESTRICTION_BUTTON
from datadoc_editor.frontend.components.identifiers import FORCE_RERENDER_COUNTER
from datadoc_editor.frontend.components.identifiers import USE_RESTRICTION_ID_STORE
from datadoc_editor.frontend.components.identifiers import (
    USE_RESTRICTION_LIST_CONTAINER,
)
from datadoc_editor.frontend.components.identifiers import USE_RESTRICTION_OPTION_STORE
from datadoc_editor.frontend.components.identifiers import USE_RESTRICTION_STORE

if TYPE_CHECKING:
    from collections.abc import Callable

    from dash.development.base_component import Component
    from pydantic import BaseModel

    from datadoc_editor.enums import LanguageStringsEnum
    from datadoc_editor.frontend.callbacks.utils import MetadataInputTypes

logger = logging.getLogger(__name__)

DATASET_METADATA_INPUT = "dataset-metadata-input"
DATASET_METADATA_DATE_INPUT = "dataset-metadata-date-input"
DATASET_METADATA_MULTILANGUAGE_INPUT = "dataset-metadata-multilanguage-input"
DATASET_METADATA_MULTIDROPDOWN_INPUT = "dataset-metadata-multidropdown-input"

PSEUDO_METADATA_INPUT = "pseudo-metadata-input"
VARIABLES_METADATA_INPUT = "variables-metadata-input"
VARIABLES_METADATA_DATE_INPUT = "variables-metadata-date-input"
VARIABLES_METADATA_MULTILANGUAGE_INPUT = "dataset-metadata-multilanguage-input"

DROPDOWN_DESELECT_OPTION = "-- Velg --"

METADATA_LANGUAGES = [
    {
        "supported_language": enums.SupportedLanguages.NORSK_BOKMÅL,
        "language_title": "Bokmål",
        "language_value": "nb",
    },
    {
        "supported_language": enums.SupportedLanguages.NORSK_NYNORSK,
        "language_title": "Nynorsk",
        "language_value": "nn",
    },
    {
        "supported_language": enums.SupportedLanguages.ENGLISH,
        "language_title": "English",
        "language_value": "en",
    },
]


def get_enum_options(
    enum: type[LanguageStringsEnum],
) -> list[dict[str, str]]:
    """Generate the list of options based on the currently chosen language."""
    dropdown_options = [
        {
            "title": i.get_value_for_language(enums.SupportedLanguages.NORSK_BOKMÅL)
            or "",
            "id": i.name,
        }
        for i in enum  # type: ignore [attr-defined]
    ]
    dropdown_options.insert(0, {"title": DROPDOWN_DESELECT_OPTION, "id": ""})
    return dropdown_options


def get_enum_options_with_delete_option(
    enum: type[LanguageStringsEnum],
) -> list[dict[str, str]]:
    """Generate the list of options based on the currently chosen language."""
    dropdown_options = [
        {
            "title": i.get_value_for_language(enums.SupportedLanguages.NORSK_BOKMÅL)
            or "",
            "id": i.name,
        }
        for i in enum  # type: ignore [attr-defined]
    ]
    dropdown_options.append({"title": DROPDOWN_DELETE_OPTION, "id": DELETE_SELECTED})
    return dropdown_options


def get_data_source_options() -> list[dict[str, str]]:
    """Collect the unit type options."""
    dropdown_options = [
        {
            "title": data_sources.get_title(enums.SupportedLanguages.NORSK_BOKMÅL),
            "id": data_sources.code,
        }
        for data_sources in state.data_sources.classifications
    ]
    dropdown_options.insert(0, {"title": DROPDOWN_DESELECT_OPTION, "id": ""})
    return dropdown_options


def get_standard_metadata(metadata: BaseModel, identifier: str) -> MetadataInputTypes:
    """Get a metadata value from the model."""
    return getattr(metadata, identifier)


def get_metadata_and_stringify(metadata: BaseModel, identifier: str) -> str | None:
    """Get a metadata value from the model and cast to string."""
    value = get_standard_metadata(metadata, identifier)
    if value is None:
        return ""
    return str(value)


def get_datetime_and_stringify(metadata: BaseModel, identifier: str) -> str | None:
    """Get a metadata datetime value from the model and cast date to string."""
    value = get_standard_metadata(metadata, identifier)
    if not value:
        return ""
    return arrow.get(str(value)).format("YYYY-MM-DD")


def _get_string_type_item(
    language_strings: model.LanguageStringType,
    current_metadata_language: enums.SupportedLanguages,
) -> str | None:
    if language_strings.root is not None:
        for i in language_strings.root:
            if i.languageCode == current_metadata_language:
                return i.languageText
    return None


def get_multi_language_metadata_and_stringify(
    metadata: BaseModel,
    identifier: str,
    language: enums.SupportedLanguages,
) -> str | None:
    """Get a metadata value supporting multiple languages from the model."""
    value: model.LanguageStringType | None = getattr(metadata, identifier)
    if value is None:
        return ""
    return _get_string_type_item(value, language)


def get_comma_separated_string(metadata: BaseModel, identifier: str) -> str:
    """Get a metadata value which is a list of strings from the model and convert it to a comma separated string."""
    value: list[str] = getattr(metadata, identifier)
    try:
        return ", ".join(value)
    except TypeError:
        # This just means we got None
        return ""


@dataclass
class DisplayMetadata(ABC):
    """Controls how a given metadata field should be displayed."""

    identifier: str
    display_name: str
    description: str
    obligatory: bool = False
    editable: bool = True

    def url_encode_shortname_ids(self, component_id: dict) -> None:
        """Encodes id to hanlde non ascii values."""
        if "variable_short_name" in component_id:
            component_id["variable_short_name"] = urllib.parse.quote(
                component_id["variable_short_name"],
            )

    @abstractmethod
    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> Component:
        """Build a component."""
        ...


@dataclass
class MetadataInputField(DisplayMetadata):
    """Controls how an input field should be displayed."""

    type: str = "text"
    value_getter: Callable[[BaseModel, str], Any] = get_metadata_and_stringify

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> ssb.Input:
        """Build an Input component."""
        self.url_encode_shortname_ids(component_id)
        return ssb.Input(
            label=self.display_name,
            id=component_id,
            debounce=True,
            type=self.type,
            showDescription=True,
            description=self.description,
            readOnly=not self.editable,
            value=self.value_getter(metadata, self.identifier),
            className="input-component",
            required=self.obligatory and self.editable,
        )


@dataclass
class MetadataDropdownField(DisplayMetadata):
    """Controls how a Dropdown should be displayed."""

    options_getter: Callable[[], list[dict[str, str]]] = list

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> ssb.Dropdown:
        """Build Dropdown component."""
        self.url_encode_shortname_ids(component_id)
        return ssb.Dropdown(
            header=self.display_name,
            id=component_id,
            items=self.options_getter(),
            placeholder=DROPDOWN_DESELECT_OPTION,
            value=get_metadata_and_stringify(metadata, self.identifier),
            className="dropdown-component",
            showDescription=True,
            description=self.description,
            required=self.obligatory and self.editable,
        )


@dataclass
class MetadataDateField(DisplayMetadata):
    """Controls how fields which define a single date are displayed."""

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> ssb.Input:
        """Build Input date component."""
        self.url_encode_shortname_ids(component_id)
        return ssb.Input(
            label=self.display_name,
            id=component_id,
            debounce=False,
            type="date",
            disabled=not self.editable,
            showDescription=True,
            description=self.description,
            value=get_metadata_and_stringify(metadata, self.identifier),
            className="input-component",
            required=self.obligatory and self.editable,
        )


@dataclass
class MetadataDateTimeField(DisplayMetadata):
    """Controls how fields which define a single date are displayed."""

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> ssb.Input:
        """Build Input date component."""
        self.url_encode_shortname_ids(component_id)
        return ssb.Input(
            label=self.display_name,
            id=component_id,
            debounce=False,
            type="date",
            disabled=not self.editable,
            showDescription=True,
            description=self.description,
            value=get_datetime_and_stringify(metadata, self.identifier),
            className="input-component",
            required=self.obligatory and self.editable,
        )


@dataclass
class MetadataPeriodField(DisplayMetadata):
    """Controls how fields which define a time period are displayed.

    These are a special case since two fields have a relationship to one another.
    """

    id_type: str = ""

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> ssb.Input:
        """Build Input date component."""
        component_id["type"] = self.id_type
        self.url_encode_shortname_ids(component_id)
        return ssb.Input(
            label=self.display_name,
            id=component_id,
            debounce=False,
            type="date",
            disabled=not self.editable,
            showDescription=True,
            description=self.description,
            value=get_metadata_and_stringify(metadata, self.identifier),
            className="input-component",
            required=self.obligatory and self.editable,
        )


@dataclass
class MetadataMultiLanguageField(DisplayMetadata):
    """Controls how fields which support multi-language are displayed.

    These are a special case since they return a group of input fields..
    """

    id_type: str = ""
    type: str = "text"

    def render_input_group(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> html.Section:
        """Build section with Input components for each language."""
        self.url_encode_shortname_ids(component_id)
        if "variable_short_name" in component_id:
            return html.Section(
                children=[
                    ssb.Input(
                        label=i["language_title"],
                        value=get_multi_language_metadata_and_stringify(
                            metadata,
                            self.identifier,
                            enums.SupportedLanguages(i["supported_language"]),
                        ),
                        debounce=True,
                        id={
                            "type": self.id_type,
                            "id": component_id["id"],
                            "variable_short_name": component_id["variable_short_name"],
                            "language": i["language_value"],
                        },
                        type=self.type,
                        className="multilanguage-input-component",
                    )
                    for i in METADATA_LANGUAGES
                ],
            )
        return html.Section(
            children=[
                ssb.Input(
                    label=i["language_title"],
                    value=get_multi_language_metadata_and_stringify(
                        metadata,
                        self.identifier,
                        enums.SupportedLanguages(i["supported_language"]),
                    ),
                    debounce=True,
                    id={
                        "type": self.id_type,
                        "id": component_id["id"],
                        "language": i["language_value"],
                    },
                    type=self.type,
                    className="multilanguage-input-component",
                )
                for i in METADATA_LANGUAGES
            ],
        )

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> html.Fieldset:
        """Build fieldset group."""
        return html.Fieldset(
            children=(
                [
                    ssb.Glossary(
                        children=(
                            html.Legend(
                                self.display_name,
                                className="multilanguage-legend",
                            ),
                            ssb.RequiredAsterisk()
                            if self.obligatory and self.editable
                            else None,
                        ),
                        explanation=self.description,
                        className="legend-glossary",
                    ),
                    self.render_input_group(
                        component_id=component_id,
                        metadata=metadata,
                    ),
                ]
            ),
            className="multilanguage-fieldset",
        )


@dataclass
class MetadataMultiDropdownField(DisplayMetadata):
    """Controls how fields which support multi-dropdown values are displayed.

    These are a special case since they return a group of both dropdown and date fields.
    This current version is only compatible with the use restriction field.
    """

    id_type: str = ""
    type: str = "text"
    options_getter: Callable[[], list[dict[str, str]]] = list
    type_display_name: str = ""
    date_display_name: str = ""
    type_description: str = ""
    date_description: str = ""

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> html.Fieldset:
        """Build fieldset group."""
        use_restrictions: list[model.UseRestrictionItem] = cast(
            "list[model.UseRestrictionItem]",
            get_standard_metadata(metadata, self.identifier) or [],
        )
        initial_data = [
            {
                "use_restriction_type": i.use_restriction_type,
                "use_restriction_date": i.use_restriction_date,
            }
            for i in use_restrictions
        ]

        idx = {"type": self.id_type, "id": component_id["id"]}

        children = [
            ssb.Glossary(
                children=[
                    html.Legend(self.display_name, className="multilanguage-legend")
                ],
                explanation=self.description,
                className="legend-glossary",
            ),
            html.Div(id=USE_RESTRICTION_LIST_CONTAINER),
            dcc.Store(id=USE_RESTRICTION_STORE, data=initial_data),
            dcc.Store(id=USE_RESTRICTION_OPTION_STORE, data=self.options_getter()),
            dcc.Store(id=USE_RESTRICTION_ID_STORE, data=idx),
            dcc.Store(id=FORCE_RERENDER_COUNTER, data=0),
            ssb.Button("Legg til bruksrestriksjon", id=ADD_USE_RESTRICTION_BUTTON),
        ]

        return html.Fieldset(children=children, className="multidropdown-fieldset")


@dataclass
class MetadataCheckboxField(DisplayMetadata):
    """Controls for how a checkbox metadata field should be displayed."""

    def render(
        self,
        component_id: dict,
        metadata: BaseModel,
    ) -> ssb.Checkbox:
        """Build Checkbox component."""
        self.url_encode_shortname_ids(component_id)
        return ssb.Checkbox(
            label=self.display_name,
            id=component_id,
            disabled=not self.editable,
            value=get_standard_metadata(metadata, self.identifier),
            showDescription=True,
            description=self.description,
            className="metadata-checkbox-field",
            required=self.obligatory and self.editable,
        )


FieldTypes = (
    MetadataInputField
    | MetadataDropdownField
    | MetadataDateField
    | MetadataDateTimeField
    | MetadataCheckboxField
    | MetadataPeriodField
    | MetadataMultiLanguageField
    | MetadataMultiDropdownField
)



"""Utility values and functions for tests."""

from pathlib import Path

TEST_BUCKET_PARQUET_FILEPATH = "gs://ssb-staging-dapla-felles-data-delt/datadoc/klargjorte_data/person_data_v1.parquet"

TEST_BUCKET_PARQUET_FILEPATH_WITH_SHORTNAME = "gs://ssb-staging-dapla-felles-data-delt/befolkning/klargjorte_data/person_data_v1.parquet"

TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH = "gs://ssb-my-team-data-produkt-prod/ifpn/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet"

TEST_RESOURCES_DIRECTORY = Path("tests/resources")
TEST_DATASETS_DIRECTORY = TEST_RESOURCES_DIRECTORY / "datasets"

TEST_PARQUET_FILE_NAME = "person_data_v1.parquet"
TEST_NAMING_STANDARD_COMPATIBLE_DATASET = (
    "ifpn/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet"
)
TEST_PARQUET_FILEPATH = TEST_DATASETS_DIRECTORY / TEST_PARQUET_FILE_NAME
TEST_SAS7BDAT_FILEPATH = TEST_DATASETS_DIRECTORY / "sasdata.sas7bdat"
TEST_PARQUET_GZIP_FILEPATH = TEST_DATASETS_DIRECTORY / "person_data_v1.parquet.gzip"

TEST_EXISTING_METADATA_DIRECTORY = TEST_RESOURCES_DIRECTORY / "existing_metadata_file"
TEST_EXISTING_METADATA_FILE_NAME = "person_data_v1__DOC.json"
TEST_EXISTING_METADATA_FILEPATH = (
    TEST_EXISTING_METADATA_DIRECTORY / TEST_EXISTING_METADATA_FILE_NAME
)
TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH = (
    TEST_EXISTING_METADATA_DIRECTORY
    / "person_testdata_p2020-12-31_p2020-12-31_v1__DOC.json"
)

TEST_EXISTING_METADATA_WITH_VALID_ID_DIRECTORY = (
    TEST_EXISTING_METADATA_DIRECTORY / "valid_id_field"
)

TEST_COMPATIBILITY_DIRECTORY = TEST_EXISTING_METADATA_DIRECTORY / "compatibility"

TEST_PROCESSED_DATA_POPULATION_DIRECTORY = TEST_RESOURCES_DIRECTORY / "klargjorte_data"


TEST_EKS_PARQUET = "test_p2021-12-31_p2021-12-31_v1.parquet"

TEST_EXISTING_METADATA_NAMING_STANDARD_FILEPATH_NO_MISSING = (
    TEST_EXISTING_METADATA_DIRECTORY / "test_p2021-12-31_p2021-12-31_v1__DOC.json"
)

TEST_ILLEGAL_DATASETS_DIRECTORY = TEST_RESOURCES_DIRECTORY / "illegal_shortnames"
TEST_PARQUET_FILE_NAME_ILLEGAL_SHORTNAMES = (
    "person_testdata_p2021-12-31_p2021-12-31_v1.parquet"
)
TEST_PARQUET_FILEPATH_ILLEGAL_SHORTNAMES = (
    TEST_ILLEGAL_DATASETS_DIRECTORY / TEST_PARQUET_FILE_NAME_ILLEGAL_SHORTNAMES
)



"""All decorated callback functions should be defined here.

Implementations of the callback functionality should be in other functions (in other files), to enable unit testing.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from typing import Any

from dash import ALL
from dash import MATCH
from dash import Dash
from dash import Input
from dash import Output
from dash import State
from dash import ctx
from dash import html
from dash import no_update

from datadoc_editor import state
from datadoc_editor.enums import PseudonymizationAlgorithmsEnum
from datadoc_editor.frontend.callbacks.dataset import accept_dataset_metadata_date_input
from datadoc_editor.frontend.callbacks.dataset import accept_dataset_metadata_input
from datadoc_editor.frontend.callbacks.dataset import accept_dataset_multidropdown_input
from datadoc_editor.frontend.callbacks.dataset import open_dataset_handling
from datadoc_editor.frontend.callbacks.dataset import remove_dataset_multidropdown_input
from datadoc_editor.frontend.callbacks.utils import render_multidropdown_row
from datadoc_editor.frontend.callbacks.utils import render_tabs
from datadoc_editor.frontend.callbacks.utils import save_metadata_and_generate_alerts
from datadoc_editor.frontend.callbacks.utils import update_store_data_with_inputs
from datadoc_editor.frontend.callbacks.variables import (
    accept_pseudo_variable_metadata_input,
)
from datadoc_editor.frontend.callbacks.variables import (
    accept_variable_metadata_date_input,
)
from datadoc_editor.frontend.callbacks.variables import accept_variable_metadata_input
from datadoc_editor.frontend.callbacks.variables import mutate_variable_pseudonymization
from datadoc_editor.frontend.callbacks.variables import populate_pseudo_workspace
from datadoc_editor.frontend.callbacks.variables import populate_variables_workspace
from datadoc_editor.frontend.components.builders import build_dataset_edit_section
from datadoc_editor.frontend.components.builders import build_dataset_machine_section
from datadoc_editor.frontend.components.identifiers import ACCORDION_WRAPPER_ID
from datadoc_editor.frontend.components.identifiers import ADD_USE_RESTRICTION_BUTTON
from datadoc_editor.frontend.components.identifiers import FORCE_RERENDER_COUNTER
from datadoc_editor.frontend.components.identifiers import SECTION_WRAPPER_ID
from datadoc_editor.frontend.components.identifiers import USE_RESTRICTION_ID_STORE
from datadoc_editor.frontend.components.identifiers import (
    USE_RESTRICTION_LIST_CONTAINER,
)
from datadoc_editor.frontend.components.identifiers import USE_RESTRICTION_OPTION_STORE
from datadoc_editor.frontend.components.identifiers import USE_RESTRICTION_STORE
from datadoc_editor.frontend.components.identifiers import VARIABLES_INFORMATION_ID
from datadoc_editor.frontend.fields.display_base import DATASET_METADATA_DATE_INPUT
from datadoc_editor.frontend.fields.display_base import DATASET_METADATA_INPUT
from datadoc_editor.frontend.fields.display_base import (
    DATASET_METADATA_MULTIDROPDOWN_INPUT,
)
from datadoc_editor.frontend.fields.display_base import (
    DATASET_METADATA_MULTILANGUAGE_INPUT,
)
from datadoc_editor.frontend.fields.display_base import PSEUDO_METADATA_INPUT
from datadoc_editor.frontend.fields.display_base import VARIABLES_METADATA_DATE_INPUT
from datadoc_editor.frontend.fields.display_base import VARIABLES_METADATA_INPUT
from datadoc_editor.frontend.fields.display_base import (
    VARIABLES_METADATA_MULTILANGUAGE_INPUT,
)
from datadoc_editor.frontend.fields.display_dataset import (
    EDITABLE_DATASET_METADATA_LEFT,
)
from datadoc_editor.frontend.fields.display_dataset import (
    EDITABLE_DATASET_METADATA_RIGHT,
)
from datadoc_editor.frontend.fields.display_dataset import NON_EDITABLE_DATASET_METADATA
from datadoc_editor.frontend.fields.display_dataset import DatasetIdentifiers
from datadoc_editor.frontend.fields.display_variables import VariableIdentifiers

if TYPE_CHECKING:
    from collections.abc import Callable

    import dash_bootstrap_components as dbc

    from datadoc_editor.frontend.callbacks.utils import MetadataInputTypes

logger = logging.getLogger(__name__)


def register_callbacks(app: Dash) -> None:  # noqa: PLR0915
    """Define and register callbacks."""

    @app.callback(
        Output("alerts-section", "children", allow_duplicate=True),
        Input("save-button", "n_clicks"),
        State("alerts-section", "children"),
        prevent_initial_call=True,
    )
    def callback_save_metadata_file(
        n_clicks: int,
        alerts: list,  # argument required by Dash  # noqa: ARG001
    ) -> Any | list:  # noqa: ANN401
        """Save the metadata document to disk and check obligatory metadata.

        Returns:
            List of alerts. Obligatory metadata alert warning if there is obligatory metadata missing.
            And success alert if metadata is saved correctly.
            If none return no_update.
        """
        if n_clicks and n_clicks > 0:
            return save_metadata_and_generate_alerts(state.metadata)

        return no_update

    @app.callback(
        Output(
            {"type": DATASET_METADATA_INPUT, "id": MATCH},
            "error",
        ),
        Output(
            {"type": DATASET_METADATA_INPUT, "id": MATCH},
            "errorMessage",
        ),
        Input(
            {"type": DATASET_METADATA_INPUT, "id": MATCH},
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
        )

    @app.callback(
        Output(
            {
                "type": DATASET_METADATA_MULTILANGUAGE_INPUT,
                "id": MATCH,
                "language": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_MULTILANGUAGE_INPUT,
                "id": MATCH,
                "language": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": DATASET_METADATA_MULTILANGUAGE_INPUT,
                "id": MATCH,
                "language": MATCH,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_multilanguage_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        # Get the ID of the input that changed. This MUST match the attribute name defined in DataDocDataSet
        return accept_dataset_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
            ctx.triggered_id["language"],
        )

    @app.callback(
        Output(
            {
                "type": DATASET_METADATA_MULTIDROPDOWN_INPUT,
                "id": MATCH,
                "field": MATCH,
                "index": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_MULTIDROPDOWN_INPUT,
                "id": MATCH,
                "field": MATCH,
                "index": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": DATASET_METADATA_MULTIDROPDOWN_INPUT,
                "id": MATCH,
                "field": MATCH,
                "index": MATCH,
            },
            "value",
        ),
    )
    def callback_accept_dataset_metadata_multidropdown_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> tuple[bool, str]:
        """Save updated dataset metadata values.

        Will display an alert if validation fails.
        """
        if ctx.triggered_id is None:
            return False, ""
        return accept_dataset_multidropdown_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["id"],
            ctx.triggered_id["field"],
            ctx.triggered_id["index"],
        )

    @app.callback(
        Output("alerts-section", "children", allow_duplicate=True),
        Output("dataset-opened-counter", "data"),  # Used to force reload of metadata
        Input("open-button", "n_clicks"),
        State("dataset-path-input", "value"),
        State("dataset-opened-counter", "data"),
        prevent_initial_call=True,
    )
    def callback_open_dataset(
        n_clicks: int,
        dataset_path: str,
        dataset_opened_counter: int,
    ) -> tuple[dbc.Alert, int]:
        """Open a dataset.

        Shows an alert on success or failure.

        To trigger reload of data in the UI, we update the
        language dropdown. This is a hack and could be replaced
        by a more formal mechanism.
        """
        return open_dataset_handling(n_clicks, dataset_path, dataset_opened_counter)

    @app.callback(
        Output("display-tab", "children"),
        Input("tabs", "value"),
    )
    def callback_render_tabs(tab: str) -> html.Article | None:
        """Return correct tab content."""
        return render_tabs(tab)

    @app.callback(
        Output(VARIABLES_INFORMATION_ID, "children"),
        Input("dataset-opened-counter", "data"),
    )
    def callback_populate_variables_info_section(
        dataset_opened_counter: int,  # noqa: ARG001 Dash requires arguments for all Inputs
    ) -> str:
        if state.metadata.variables and len(state.metadata.variables) > 0:
            return f"Datasettet inneholder {len(state.metadata.variables)} variabler."

        return "Åpne et datasett for å liste variablene."

    @app.callback(
        Output(ACCORDION_WRAPPER_ID, "children"),
        Input("dataset-opened-counter", "data"),
        Input("search-variables", "value"),
    )
    def callback_populate_variables_workspace(
        dataset_opened_counter: int,
        search_query: str,
    ) -> list:
        """Create variable workspace with accordions for variables."""
        logger.debug("Populating variables workspace. Search query: %s", search_query)
        return populate_variables_workspace(
            state.metadata.variables,
            search_query,
            dataset_opened_counter,
        )

    @app.callback(
        [Output(USE_RESTRICTION_STORE, "data"), Output(FORCE_RERENDER_COUNTER, "data")],
        [
            Input(ADD_USE_RESTRICTION_BUTTON, "n_clicks"),
            Input(
                {
                    "type": DATASET_METADATA_MULTIDROPDOWN_INPUT,
                    "id": ALL,
                    "index": ALL,
                    "field": "type",
                },
                "value",
            ),
            Input(
                {
                    "type": DATASET_METADATA_MULTIDROPDOWN_INPUT,
                    "id": ALL,
                    "index": ALL,
                    "field": "date",
                },
                "value",
            ),
            Input(
                {
                    "type": DATASET_METADATA_MULTIDROPDOWN_INPUT,
                    "id": ALL,
                    "index": ALL,
                    "field": "delete",
                },
                "n_clicks",
            ),
        ],
        [State(USE_RESTRICTION_STORE, "data"), State(FORCE_RERENDER_COUNTER, "data")],
        prevent_initial_call=True,
    )
    def handle_add_and_delete(  # noqa: PLR0913
        add_clicks: int,  # noqa: ARG001
        type_values: list[str],
        date_values: list[str],
        delete_clicks: list[int],  # noqa: ARG001
        store_data: list | None,
        counter: int,
    ) -> tuple[list, int]:
        triggered = ctx.triggered_id
        counter += 1
        store_data = update_store_data_with_inputs(
            store_data or [], type_values, date_values
        )

        if triggered == ADD_USE_RESTRICTION_BUTTON:
            store_data.append(
                {"use_restriction_type": None, "use_restriction_date": None}
            )

        elif isinstance(triggered, dict) and triggered.get("field") == "delete":
            idx = triggered.get("index")
            if isinstance(idx, int) and 0 <= idx < len(store_data):
                remove_dataset_multidropdown_input(
                    DATASET_METADATA_MULTIDROPDOWN_INPUT, idx
                )
                store_data.pop(idx)

        return store_data, counter

    @app.callback(
        Output(USE_RESTRICTION_LIST_CONTAINER, "children"),
        Input(USE_RESTRICTION_STORE, "data"),
        Input(USE_RESTRICTION_OPTION_STORE, "data"),
        Input(USE_RESTRICTION_ID_STORE, "data"),
        Input(FORCE_RERENDER_COUNTER, "data"),
    )
    def render_use_restriction_list(
        current_list: list,
        options: Callable[[], list[dict[str, str]]],
        idx: dict[str, str | int],
        rerender_counter: int,
    ) -> list:
        if not current_list:
            return []

        items = []
        for i, item in enumerate(current_list):
            row_id = {**idx, "index": i}
            items.append(
                render_multidropdown_row(
                    item, row_id, options, key=f"{rerender_counter}-{i}"
                )
            )
        return items

    @app.callback(
        Output(SECTION_WRAPPER_ID, "children"),
        Input("dataset-opened-counter", "data"),
    )
    def callback_populate_dataset_workspace(
        dataset_opened_counter: int,  # Dash requires arguments for all Inputs
    ) -> list:
        """Create dataset workspace with sections."""
        logger.debug("Populating dataset workspace")
        return [
            build_dataset_edit_section(
                [
                    EDITABLE_DATASET_METADATA_LEFT,
                    EDITABLE_DATASET_METADATA_RIGHT,
                ],
                state.metadata.dataset,
                {
                    "type": "dataset-edit-section",
                    "id": f"obligatory-{dataset_opened_counter}",
                },
            ),
            build_dataset_machine_section(
                "Maskingenerert",
                NON_EDITABLE_DATASET_METADATA,
                state.metadata.dataset,
                {
                    "type": "dataset-machine-section",
                    "id": f"machine-{dataset_opened_counter}",
                },
            ),
        ]

    @app.callback(
        Output(
            {"type": PSEUDO_METADATA_INPUT, "variable_short_name": MATCH, "id": MATCH},
            "error",
        ),
        Output(
            {"type": PSEUDO_METADATA_INPUT, "variable_short_name": MATCH, "id": MATCH},
            "errorMessage",
        ),
        Input(
            {"type": PSEUDO_METADATA_INPUT, "variable_short_name": MATCH, "id": MATCH},
            "value",
        ),
        State(
            {"type": PSEUDO_METADATA_INPUT, "variable_short_name": MATCH, "id": MATCH},
            "id",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_pseudo_variable_metadata_input(value, component_id):  # noqa: ANN202, ANN001
        if value is None or component_id is None:
            # Nothing to do if deselected or missing
            return False, ""
        variable_short_name = component_id["variable_short_name"]
        input_id = component_id["id"]
        logger.debug(
            "Callback triggered with value=%s, component_id=%s",
            value,
            component_id,
        )
        # Safely get variable from state
        variable = state.metadata.variables_lookup.get(variable_short_name)
        if not variable:
            logger.info("Variable not found: %s", variable_short_name)
            return False, "Variable not found."

        message = accept_pseudo_variable_metadata_input(
            value, variable_short_name, input_id
        )

        if not message:
            return False, ""

        return True, message

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_MULTILANGUAGE_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
                "language": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_MULTILANGUAGE_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
                "language": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_MULTILANGUAGE_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
                "language": MATCH,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_multilanguage_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> dbc.Alert:
        """Save updated variable metadata values."""
        message = accept_variable_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["variable_short_name"],
            ctx.triggered_id["id"],
            ctx.triggered_id["language"],
        )
        if not message:
            # No error to display.
            return False, ""

        return True, message

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "errorMessage",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "value",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_DATE_INPUT,
                "variable_short_name": MATCH,
                "id": VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_date_input(
        contains_data_from: str,
        contains_data_until: str,
    ) -> dbc.Alert:
        """Special case handling for date fields which have a relationship to one another."""
        return accept_variable_metadata_date_input(
            VariableIdentifiers(ctx.triggered_id["id"]),
            ctx.triggered_id["variable_short_name"],
            contains_data_from,
            contains_data_until,
        )

    @app.callback(
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "errorMessage",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "error",
        ),
        Output(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_FROM.value,
            },
            "value",
        ),
        Input(
            {
                "type": DATASET_METADATA_DATE_INPUT,
                "id": DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_dataset_metadata_date_input(
        contains_data_from: str,
        contains_data_until: str,
    ) -> dbc.Alert:
        """Special case handling for date fields which have a relationship to one another."""
        return accept_dataset_metadata_date_input(
            DatasetIdentifiers(ctx.triggered_id["id"]),
            contains_data_from,
            contains_data_until,
        )

    @app.callback(
        Output(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "error",
        ),
        Output(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "errorMessage",
        ),
        Input(
            {
                "type": VARIABLES_METADATA_INPUT,
                "variable_short_name": MATCH,
                "id": MATCH,
            },
            "value",
        ),
        prevent_initial_call=True,
    )
    def callback_accept_variable_metadata_input(
        value: MetadataInputTypes,  # noqa: ARG001 argument required by Dash
    ) -> dbc.Alert:
        """Save updated variable metadata values."""
        message = accept_variable_metadata_input(
            ctx.triggered[0]["value"],
            ctx.triggered_id["variable_short_name"],
            ctx.triggered_id["id"],
        )
        if not message:
            # No error to display.
            return False, ""

        return True, message

    @app.callback(
        Output({"type": "pseudo-field-container", "variable": MATCH}, "children"),
        Input({"type": "pseudonymization-dropdown", "variable": MATCH}, "value"),
        Input("save-button", "n_clicks"),
        State({"type": "pseudonymization-dropdown", "variable": MATCH}, "id"),
    )
    def callback_populate_pseudo_workspace(
        value,  # noqa: ANN001
        n_clicks: int,
        dropdown_id,  # noqa: ANN001
    ) -> dbc.Form:
        """Dynamically create pseudonymization workspace.

        - The dropdown value updates the displayed pseudonymization fields immediately.
        - The Save button applies permanent changes (update or deletion) to the variable.
        """
        # Map dropdown value to enum if possible
        selected_algorithm = (
            PseudonymizationAlgorithmsEnum[value]
            if value and value in PseudonymizationAlgorithmsEnum.__members__
            else value
        )

        logger.debug("Selected algorithm: %s", selected_algorithm)
        variable = state.metadata.variables_lookup.get(dropdown_id["variable"])

        if variable is None:
            logger.info("Variable not found in lookup!")
            return []

        # Persist update and deletion only on save
        if n_clicks and n_clicks > 0:
            mutate_variable_pseudonymization(variable, selected_algorithm)

        logger.debug(
            "Variable %s has pseudo info: %s",
            variable.short_name,
            variable.pseudonymization,
        )
        return populate_pseudo_workspace(variable, selected_algorithm)



"""Dash callback functions.

The actual decorated callbacks should be very minimal functions which
call other functions where the logic lives. This is done to support unit
testing because the decorated functions are difficult to test.

The functions where the logic lives should be categorised into files.
"""



"""Test function build_dataset_edit_section."""

import pytest
import ssb_dash_components as ssb  # type: ignore[import-untyped]
from dapla_metadata.datasets import model
from dash import html

from datadoc_editor.frontend.components.builders import build_dataset_edit_section
from datadoc_editor.frontend.fields.display_dataset import DISPLAY_DATASET
from datadoc_editor.frontend.fields.display_dataset import (
    EDITABLE_DATASET_METADATA_LEFT,
)
from datadoc_editor.frontend.fields.display_dataset import (
    EDITABLE_DATASET_METADATA_RIGHT,
)
from datadoc_editor.frontend.fields.display_dataset import NON_EDITABLE_DATASET_METADATA
from datadoc_editor.frontend.fields.display_dataset import DatasetIdentifiers

# List of obligatory and editable fields minus dropdowns with atypical options-getters
NON_ATYPICAL_DROPDOWNS = [
    m
    for m in DISPLAY_DATASET.values()
    if m.editable
    and m.identifier
    not in (
        DatasetIdentifiers.SUBJECT_FIELD.value,
        DatasetIdentifiers.OWNER.value,
    )
]

INPUT_DATA_BUILD_DATASET_SECTION = [
    (
        [NON_EDITABLE_DATASET_METADATA, []],
        model.Dataset(short_name="fantastic_dataset"),
        {"type": "dataset-edit-section", "id": "obligatory-nb"},
    ),
    (
        [[], []],
        model.Dataset(short_name="fantastic_dataset"),
        {"type": "dataset-edit-section", "id": "obligatory-nb"},
    ),
    (
        [[], []],
        None,
        {},
    ),
    (
        [[], []],
        None,
        {"type": "dataset-edit-section", "id": "obligatory-en"},
    ),
    (
        {},
        model.Dataset(short_name="last_dataset"),
        {},
    ),
    (
        [NON_ATYPICAL_DROPDOWNS, []],
        model.Dataset(),
        {"type": "dataset-edit-section", "id": "obligatory-nb"},
    ),
]


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_types", "dataset", "key"),
    INPUT_DATA_BUILD_DATASET_SECTION,
)
def test_build_dataset_section_is_html_section(
    field_types,
    dataset,
    key,
):
    edit_section = build_dataset_edit_section(
        field_types,
        dataset,
        key,
    )
    assert isinstance(edit_section, html.Section)
    assert edit_section._namespace == "dash_html_components"  # noqa: SLF001
    assert edit_section.id == key


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.usefixtures("_statistic_subject_mapping_fake_subjects")
@pytest.mark.parametrize(
    (
        "edit_section",
        "expected_fields_left",
        "expected_fields_right",
    ),
    [
        (
            (
                [EDITABLE_DATASET_METADATA_LEFT, EDITABLE_DATASET_METADATA_RIGHT],
                model.Dataset(short_name="dataset"),
                {"type": "dataset-edit-section", "id": "dataset-nb"},
            ),
            6,
            8,
        ),
    ],
)
def test_build_dataset_edit_section_renders_ssb_components(
    edit_section,
    expected_fields_left,
    expected_fields_right,
):
    edit_section = build_dataset_edit_section(*edit_section)
    fields_left = edit_section.children[0].children
    fields_right = edit_section.children[1].children

    assert len(fields_left) == expected_fields_left
    assert len(fields_right) == expected_fields_right


NEW_INPUT_DATA_BUILD_DATASET_SECTION = [
    (
        [EDITABLE_DATASET_METADATA_LEFT, EDITABLE_DATASET_METADATA_RIGHT],
        model.Dataset(short_name="dataset"),
        {"type": "dataset-edit-section", "id": "obligatory-nb"},
    ),
    (
        [[], []],
        model.Dataset(short_name="dataset"),
        {"type": "dataset-edit-section", "id": "obligatory-nb"},
    ),
    (
        [[], []],
        None,
        {},
    ),
    (
        [[], []],
        None,
        {"type": "dataset-edit-section", "id": "obligatory-en"},
    ),
    (
        [[], []],
        model.Dataset(short_name="last_dataset"),
        {},
    ),
]


@pytest.mark.parametrize(
    ("field_types", "dataset", "key"),
    NEW_INPUT_DATA_BUILD_DATASET_SECTION,
)
@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.usefixtures("_statistic_subject_mapping_fake_subjects")
def test_build_dataset_edit_section_input_component_props(
    field_types,
    dataset,
    key,
):
    edit_section = build_dataset_edit_section(
        field_types,
        dataset,
        key,
    )
    fields = edit_section.children[0].children + edit_section.children[1].children

    input_components = [element for element in fields if isinstance(element, ssb.Input)]

    assert all(
        item.type in ("text", "url", "date", "number") for item in input_components
    )
    assert all(
        (
            item.debounce is True
            if item.type in ("text", "url")
            else item.debounce is False
        )
        for item in input_components
        if item.type != "number"
    )
    assert all(item.label != "" for item in input_components)


@pytest.mark.parametrize(
    ("field_types", "dataset", "key"),
    NEW_INPUT_DATA_BUILD_DATASET_SECTION,
)
@pytest.mark.usefixtures(
    "_code_list_fake_classifications",
    "_statistic_subject_mapping_fake_subjects",
)
def test_build_dataset_edit_section_dropdown_component_props(field_types, dataset, key):
    edit_section = build_dataset_edit_section(field_types, dataset, key)

    fields_left = edit_section.children[0].children
    fields_right = edit_section.children[1].children

    for fields in (fields_left, fields_right):
        dropdown_components = [
            element for element in fields if isinstance(element, ssb.Dropdown)
        ]
        assert all(item.items for item in dropdown_components)
        assert all(item.header for item in dropdown_components)



"""Shared fixtures and configuration."""

from __future__ import annotations

import concurrent
import copy
import functools
import logging
import os
import pathlib
import shutil
from datetime import UTC
from datetime import datetime
from typing import TYPE_CHECKING

import pandas as pd
import pytest
from bs4 import BeautifulSoup
from bs4 import ResultSet
from dapla_metadata.dapla.user_info import TestUserInfo  # type:ignore [import-untyped]
from dapla_metadata.datasets import Datadoc
from dapla_metadata.datasets import model
from dapla_metadata.datasets.code_list import CodeList
from dapla_metadata.datasets.statistic_subject_mapping import StatisticSubjectMapping

from datadoc_editor import state

from .utils import TEST_EXISTING_METADATA_DIRECTORY
from .utils import TEST_PARQUET_FILE_NAME
from .utils import TEST_PARQUET_FILE_NAME_ILLEGAL_SHORTNAMES
from .utils import TEST_PARQUET_FILEPATH
from .utils import TEST_PARQUET_FILEPATH_ILLEGAL_SHORTNAMES
from .utils import TEST_RESOURCES_DIRECTORY

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture

logging.getLogger("faker").setLevel(logging.ERROR)

DATADOC_METADATA_MODULE = "dapla_metadata.datasets"
CODE_LIST_DIR = "code_list"
STATISTICAL_SUBJECT_STRUCTURE_DIR = "statistical_subject_structure"


@pytest.fixture(autouse=True)
def _clear_environment(mocker: MockerFixture) -> None:
    """Ensure that the environment is cleared."""
    mocker.patch.dict(os.environ, clear=True)


@pytest.fixture(scope="session", autouse=True)
def faker_session_locale():
    return ["no_NO"]


@pytest.fixture
def dummy_timestamp() -> datetime:
    return datetime(2022, 1, 1, tzinfo=UTC)


@pytest.fixture
def _mock_timestamp(mocker: MockerFixture, dummy_timestamp: datetime) -> None:
    mocker.patch(
        DATADOC_METADATA_MODULE + ".core.get_timestamp_now",
        return_value=dummy_timestamp,
    )


@pytest.fixture
def _mock_user_info(mocker: MockerFixture) -> None:
    mocker.patch(
        "dapla_metadata.dapla.user_info.get_user_info_for_current_platform",
        return_value=TestUserInfo(),
    )


@pytest.fixture
def metadata(
    _mock_timestamp: None,
    _mock_user_info: None,
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: Path,
) -> Datadoc:
    shutil.copy(TEST_PARQUET_FILEPATH, tmp_path / TEST_PARQUET_FILE_NAME)
    return Datadoc(
        str(tmp_path / TEST_PARQUET_FILE_NAME),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
    )


@pytest.fixture
def metadata_illegal_shortnames(
    _mock_timestamp: None,
    _mock_user_info: None,
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    tmp_path: Path,
) -> Datadoc:
    shutil.copy(
        TEST_PARQUET_FILEPATH_ILLEGAL_SHORTNAMES,
        tmp_path / TEST_PARQUET_FILE_NAME_ILLEGAL_SHORTNAMES,
    )
    return Datadoc(
        str(tmp_path / TEST_PARQUET_FILE_NAME_ILLEGAL_SHORTNAMES),
        statistic_subject_mapping=subject_mapping_fake_statistical_structure,
    )


@pytest.fixture
def existing_metadata_path() -> Path:
    return TEST_EXISTING_METADATA_DIRECTORY


@pytest.fixture(autouse=True)
def _clear_state() -> None:
    """Global fixture, referred to in pytest.ini."""
    try:
        del state.metadata
        del state.statistic_subject_mapping
    except AttributeError:
        pass


@pytest.fixture
def english_name() -> str:
    return "English Name"


@pytest.fixture
def bokmål_name() -> str:
    return "Bokmål navn"


@pytest.fixture
def nynorsk_name() -> str:
    return "Nynorsk namn"


@pytest.fixture
def language_object(
    english_name: str,
    bokmål_name: str,
    nynorsk_name: str,
) -> model.LanguageStringType:
    return model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="en", languageText=english_name),
            model.LanguageStringTypeItem(languageCode="nb", languageText=bokmål_name),
            model.LanguageStringTypeItem(languageCode="nn", languageText=nynorsk_name),
        ],
    )


@pytest.fixture
def full_dataset_state_path(
    path_parts_to_insert: str | list[str],
) -> pathlib.Path:
    """Create a longer path structure from just one section.

    Examples:
    >>> full_dataset_state_path('inndata')
    'tests/inndata/resources/person_data_v1.parquet'
    >>> full_dataset_state_path(['klargjorte_data', 'arbmark'])
    'tests/klargjorte_data/arbmark/resources/person_data_v1.parquet'
    """
    split_path = list(pathlib.Path(TEST_PARQUET_FILEPATH).parts)
    new_path = copy.copy(split_path)

    if isinstance(path_parts_to_insert, str):
        parts = [path_parts_to_insert]
    else:
        parts = path_parts_to_insert
    for p in parts:
        new_path.insert(-2, p)
    return pathlib.Path().joinpath(*new_path)


@pytest.fixture
def subject_xml_file_path() -> pathlib.Path:
    return (
        TEST_RESOURCES_DIRECTORY
        / STATISTICAL_SUBJECT_STRUCTURE_DIR
        / "extract_secondary_subject.xml"
    )


@pytest.fixture
def thread_pool_executor() -> concurrent.futures.ThreadPoolExecutor:
    return concurrent.futures.ThreadPoolExecutor(max_workers=12)


@pytest.fixture
def subject_mapping_fake_statistical_structure(
    _mock_fetch_statistical_structure,
    thread_pool_executor,
) -> StatisticSubjectMapping:
    return StatisticSubjectMapping(thread_pool_executor, "placeholder")


@pytest.fixture
def _mock_fetch_statistical_structure(
    mocker,
    subject_xml_file_path: pathlib.Path,
) -> None:
    def fake_statistical_structure() -> ResultSet:
        with subject_xml_file_path.open() as f:
            return BeautifulSoup(f.read(), features="xml").find_all("hovedemne")

    mocker.patch(
        DATADOC_METADATA_MODULE
        + ".statistic_subject_mapping.StatisticSubjectMapping._fetch_data_from_external_source",
        functools.partial(fake_statistical_structure),
    )


@pytest.fixture
def _statistic_subject_mapping_fake_subjects(
    subject_mapping_fake_statistical_structure,
) -> None:
    state.statistic_subject_mapping = subject_mapping_fake_statistical_structure
    state.statistic_subject_mapping.wait_for_external_result()


@pytest.fixture
def subject_mapping_http_exception(
    requests_mock,
    exception_to_raise,
    thread_pool_executor,
) -> StatisticSubjectMapping:
    requests_mock.get(
        "http://test.some.url.com",
        exc=exception_to_raise,
    )
    return StatisticSubjectMapping(thread_pool_executor, "http://test.some.url.com")


@pytest.fixture
def code_list_csv_filepath_nb() -> pathlib.Path:
    return TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_nb.csv"


@pytest.fixture
def code_list_csv_filepath_nn() -> pathlib.Path:
    return TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_nn.csv"


@pytest.fixture
def code_list_csv_filepath_en() -> pathlib.Path:
    return TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_en.csv"


@pytest.fixture
def _mock_fetch_dataframe(
    mocker,
    code_list_csv_filepath_nb: pathlib.Path,
    code_list_csv_filepath_nn: pathlib.Path,
    code_list_csv_filepath_en: pathlib.Path,
) -> None:
    def fake_code_list() -> dict[str, pd.DataFrame]:
        return {
            "nb": pd.read_csv(code_list_csv_filepath_nb, converters={"code": str}),
            "nn": pd.read_csv(code_list_csv_filepath_nn, converters={"code": str}),
            "en": pd.read_csv(code_list_csv_filepath_en, converters={"code": str}),
        }

    mocker.patch(
        DATADOC_METADATA_MODULE
        + ".code_list.CodeList._fetch_data_from_external_source",
        functools.partial(fake_code_list),
    )


@pytest.fixture
def code_list_fake_structure(_mock_fetch_dataframe, thread_pool_executor) -> CodeList:
    return CodeList(thread_pool_executor, 100)


@pytest.fixture
def _code_list_fake_classifications(code_list_fake_structure) -> None:
    state.measurement_units = code_list_fake_structure
    state.measurement_units.wait_for_external_result()

    state.data_sources = code_list_fake_structure
    state.data_sources.wait_for_external_result()

    state.unit_types = code_list_fake_structure
    state.unit_types.wait_for_external_result()

    state.organisational_units = code_list_fake_structure
    state.organisational_units.wait_for_external_result()


@pytest.fixture
def copy_dataset_to_path(
    tmp_path: pathlib.Path,
    full_dataset_state_path: pathlib.Path,
) -> pathlib.Path:
    temporary_dataset = tmp_path / full_dataset_state_path
    temporary_dataset.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(TEST_PARQUET_FILEPATH, temporary_dataset)
    return temporary_dataset



"""Test function build_edit_section."""

import dash_bootstrap_components as dbc
import pytest
from dapla_metadata.datasets import model
from dash import html

from datadoc_editor.frontend.components.builders import build_edit_section
from datadoc_editor.frontend.fields.display_variables import VARIABLES_METADATA_LEFT
from datadoc_editor.frontend.fields.display_variables import VARIABLES_METADATA_RIGHT


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    [
        ([[], []], model.Variable()),
        (
            [VARIABLES_METADATA_LEFT, VARIABLES_METADATA_RIGHT],
            model.Variable(short_name="pers_id"),
        ),
    ],
)
def test_build_edit_section_return_correct_component(
    field_list,
    variable,
):
    """Assert method returns html section."""
    edit_section = build_edit_section(field_list, variable)
    assert isinstance(edit_section, html.Section)


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    [
        (
            [[], []],
            model.Variable(),
        ),
        (
            [VARIABLES_METADATA_LEFT, VARIABLES_METADATA_RIGHT],
            model.Variable(short_name="sivilstand"),
        ),
    ],
)
def test_build_edit_section_children_return_correct_components(
    field_list,
    variable,
):
    """Assert method has a list of children which returns SSB dash component and dash bootstrap Form."""
    edit_section = build_edit_section(field_list, variable)
    form = edit_section.children[1]
    assert isinstance(form, dbc.Form)


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    [
        ([[], []], model.Variable()),
        (
            [VARIABLES_METADATA_LEFT, VARIABLES_METADATA_RIGHT],
            model.Variable(short_name="sykepenger"),
        ),
    ],
)
def test_build_edit_section_has_correct_id(field_list, variable):
    """Assert dictionary id is build correctly with 'type' and."""
    edit_section = build_edit_section(field_list, variable)
    assert edit_section.id == {"type": "edit-section"}



"""Smoke tests."""

from datadoc_editor import state
from datadoc_editor.app import get_app


def test_get_app(
    subject_mapping_fake_statistical_structure,
    code_list_fake_structure,
    thread_pool_executor,
):
    state.statistic_subject_mapping = subject_mapping_fake_statistical_structure
    state.code_list = code_list_fake_structure

    app, _ = get_app(thread_pool_executor)
    assert app.config["name"] == "Datadoc"
    assert len(app.callback_map.items()) > 0



from enum import Enum

from datadoc_editor.frontend.fields.display_base import FieldTypes
from datadoc_editor.frontend.fields.display_base import MetadataDateTimeField
from datadoc_editor.frontend.fields.display_base import MetadataInputField


class PseudoVariableIdentifiers(str, Enum):
    """Pseudo fields."""

    PSEUDONYMIZATION_TIME = "pseudonymization_time"
    STABLE_IDENTIFIER_TYPE = "stable_identifier_type"
    STABLE_IDENTIFIER_VERSION = "stable_identifier_version"
    ENCRYPTION_ALGORITHM = "encryption_algorithm"
    ENCRYPTION_KEY_REFERENCE = "encryption_key_reference"
    ENCRYPTION_ALGORITHM_PARAMETERS = "encryption_algorithm_parameters"


PSEUDO_FIELDS: dict[
    PseudoVariableIdentifiers,
    FieldTypes,
] = {
    PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME: MetadataDateTimeField(
        identifier=PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME.value,
        display_name="Pseudonymiseringstidspunkt",
        description="Tidspunktet datasettet ble pseudonymisert.",
        obligatory=False,
    ),
    PseudoVariableIdentifiers.STABLE_IDENTIFIER_VERSION: MetadataInputField(
        identifier=PseudoVariableIdentifiers.STABLE_IDENTIFIER_VERSION.value,
        display_name="Stabil identifikator versjon",
        description="Det skal brukes den type versjonering som er brukt for stabil identifikator katalogen, eksempler kan være dato eller semantisk versjonering.",
        obligatory=False,
    ),
    PseudoVariableIdentifiers.STABLE_IDENTIFIER_TYPE: MetadataInputField(
        identifier=PseudoVariableIdentifiers.STABLE_IDENTIFIER_TYPE.value,
        display_name="Stabil identifikator type",
        description="Type stabil identifikator som er benyttet før pseudonymisering (kryptering), eksempelvis at fødselsnummer er byttet ut med SNR fra SNR-katalogen i FREG.",
        obligatory=False,
    ),
    PseudoVariableIdentifiers.ENCRYPTION_ALGORITHM: MetadataInputField(
        identifier=PseudoVariableIdentifiers.ENCRYPTION_ALGORITHM.value,
        display_name="Krypteringsalgoritme",
        description="Krypteringsalgoritmen som er benyttet for å pseudonymisere variabelen.",
        obligatory=True,
    ),
    PseudoVariableIdentifiers.ENCRYPTION_KEY_REFERENCE: MetadataInputField(
        identifier=PseudoVariableIdentifiers.ENCRYPTION_KEY_REFERENCE.value,
        display_name="Krypteringsnøkkel referanse",
        description="Navn eller referanse til krypteringsnøkkelen som er benyttet til å pseudonymisere variabelen.",
        obligatory=True,
    ),
    PseudoVariableIdentifiers.ENCRYPTION_ALGORITHM_PARAMETERS: MetadataInputField(
        identifier=PseudoVariableIdentifiers.ENCRYPTION_ALGORITHM_PARAMETERS.value,
        display_name="Krypteringsalgoritme-parametere",
        description="Eventuelle krypteringsalgoritme-parametere som er benyttet utover “encryption_key_reference” nevnt over.",
        obligatory=False,
        editable=False,
    ),
}

PSEUDONYMIZATION_METADATA = [m for m in PSEUDO_FIELDS.values() if m.editable]
PSEUDONYMIZATION_PAPIS_WITH_STABLE_ID_METADATA = [
    m
    for m in PSEUDO_FIELDS.values()
    if (
        m.identifier
        in (
            PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME.value,
            PseudoVariableIdentifiers.STABLE_IDENTIFIER_VERSION.value,
        )
        and m.editable
    )
]
PSEUDONYMIZATION_PAPIS_WITHOUT_STABLE_ID_METADATA = [
    m
    for m in PSEUDO_FIELDS.values()
    if (
        m.identifier in (PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME.value)
        and m.editable
    )
]
PSEUDONYMIZATION_DEAD_METADATA = [
    m
    for m in PSEUDO_FIELDS.values()
    if (
        m.identifier in (PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME.value)
        and m.editable
    )
]

OBLIGATORY_VARIABLES_METADATA_PSEUDO_IDENTIFIERS_AND_DISPLAY_NAME: list[tuple] = [
    (m.identifier, m.display_name) for m in PSEUDO_FIELDS.values() if m.obligatory
]



"""Repository for constant values in Datadoc."""

MISSING_METADATA_WARNING = "Advarsel - obligatorisk metadata mangler"
CHECK_OBLIGATORY_METADATA_DATASET_MESSAGE = "Følgende datasett felt har ikke verdi:"
CHECK_OBLIGATORY_METADATA_VARIABLES_MESSAGE = "Følgende variabler har felt uten verdi:"
DAPLA_MANUAL_TEXT = "Dapla manual navnestandard"

ILLEGAL_SHORTNAME_WARNING = (
    "Noen av variablene i datasetter følger ikke navnestandard for kortnavn"
)
ILLEGAL_SHORTNAME_WARNING_MESSAGE = "Følgende navnestandard er utarbeidet for variabler: Alfanumerisk begrenset til a-z (kun små bokstaver), 0-9 og _ (understrek). Kortnavn som ikke følger standarden:"

STANDARD_ALGORITM_DAPLA_ENCRYPTION = "TINK-DAEAD"
PAPIS_ALGORITHM_ENCRYPTION = "TINK-FPE"
PAPIS_STABLE_IDENTIFIER_TYPE = "FREG_SNR"
PAPIS_ENCRYPTION_KEY_REFERENCE = "papis-common-key-1"
DAEAD_ENCRYPTION_KEY_REFERENCE = "ssb-common-key-1"
ENCRYPTION_PARAMETER_SNAPSHOT_DATE = "snapshotDate"
ENCRYPTION_PARAMETER_KEY_ID = "keyId"
ENCRYPTION_PARAMETER_STRATEGY = "strategy"
ENCRYPTION_PARAMETER_STRATEGY_SKIP = "skip"

DELETE_SELECTED = "delete_selected"
DROPDOWN_DELETE_OPTION = "Ingen (slett)"



from __future__ import annotations

from typing import Any

from datadoc_editor.config import get_log_formatter
from datadoc_editor.config import get_log_level
from datadoc_editor.logging_configuration.gunicorn_access_log_filter import (
    GunicornAccessLoggerHealthProbeFilter,
)


def get_log_config() -> dict[str, Any]:
    """Configure logging for the application."""
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {
                "format": "%(asctime)s.%(msecs)03d %(levelname)s %(name)s: %(message)s",
                "logger": "name",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
            "json": {
                "()": "datadoc_editor.logging_configuration.json_formatter.DatadocJSONFormatter",
                "fmt_keys": {
                    "level": "levelname",
                    "message": "message",
                    "timestamp": "timestamp",
                    "logger": "name",
                    "module": "module",
                    "function": "funcName",
                    "line": "lineno",
                    "thread_name": "threadName",
                },
            },
        },
        "handlers": {
            "stdout": {
                "class": "logging.StreamHandler",
                "level": get_log_level(),
                "formatter": get_log_formatter(),
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "gunicorn": {
                "handlers": ["stdout"],
                "level": "INFO",
                "propagate": False,
            },
            "gunicorn.access": {
                "handlers": ["stdout"],
                "level": "INFO",
                "propagate": False,
                "filters": [GunicornAccessLoggerHealthProbeFilter()],
            },
            "gunicorn.error": {
                "handlers": ["stdout"],
                "level": "INFO",
                "propagate": False,
            },
        },
        "root": {
            "level": get_log_level(),
            "handlers": [
                "stdout",
            ],
        },
    }



import logging

EXCLUDED_PATHS = ["/healthz", "/_dash-", "/assets"]


class GunicornAccessLoggerHealthProbeFilter(logging.Filter):
    """Filter out any Gunicorn access logs on Liveness or Readiness probes."""

    def filter(self, record: logging.LogRecord) -> bool:
        """Filter health probes on the /healthz endpoints."""
        return all(path not in record.getMessage() for path in EXCLUDED_PATHS)



"""Tests for the utils module."""

import pathlib

import tomli

from datadoc_editor.utils import get_app_version
from datadoc_editor.utils import running_in_notebook


def test_not_running_in_notebook():
    assert not running_in_notebook()


def test_get_app_version():
    with (pathlib.Path(__file__).parent.parent / "pyproject.toml").open("rb") as f:
        pyproject = tomli.load(f)

    assert get_app_version() == pyproject["project"]["version"]



"""Test function build_input_field_section."""

import dash_bootstrap_components as dbc
import pytest
import ssb_dash_components as ssb  # type: ignore[import-untyped]
from dapla_metadata.datasets import model

from datadoc_editor.frontend.components.builders import build_input_field_section
from datadoc_editor.frontend.fields.display_base import MetadataCheckboxField
from datadoc_editor.frontend.fields.display_base import MetadataDropdownField
from datadoc_editor.frontend.fields.display_base import MetadataInputField
from datadoc_editor.frontend.fields.display_base import MetadataPeriodField
from datadoc_editor.frontend.fields.display_variables import VARIABLES_METADATA_LEFT
from datadoc_editor.frontend.fields.display_variables import VARIABLES_METADATA_RIGHT

VARIABLES_METADATA = VARIABLES_METADATA_LEFT + VARIABLES_METADATA_RIGHT

INPUT_FIELD_SECTION = [
    (
        VARIABLES_METADATA,
        model.Variable(short_name="hoveddiagnose"),
    ),
    (
        VARIABLES_METADATA,
        model.Variable(short_name="pers_id"),
    ),
    (
        VARIABLES_METADATA,
        model.Variable(short_name="ber_bruttoformue"),
    ),
    (
        VARIABLES_METADATA,
        model.Variable(short_name="sykepenger"),
    ),
]


@pytest.mark.usefixtures("_code_list_fake_classifications")
def test_build_input_field_section_no_input():
    """Assert build with empty inputs returns dash bootstrap Form component and empty children list."""
    input_section = build_input_field_section(
        [],
        "",
        "",
    )
    assert input_section.children == []
    assert isinstance(input_section, dbc.Form)


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    INPUT_FIELD_SECTION,
)
def test_build_input_fields_input_components(
    field_list,
    variable,
):
    input_section = build_input_field_section(field_list, "left", variable)

    elements_of_input = [
        element for element in input_section.children if isinstance(element, ssb.Input)
    ]

    elements_of_input_and_type_text_url = [
        element
        for element in elements_of_input
        if element.type in {"text", "url", "number"}
    ]

    variable_identifier_input = [
        element
        for element in VARIABLES_METADATA
        if isinstance(element, MetadataInputField)
    ]

    assert all(isinstance(field, ssb.Input) for field in elements_of_input)
    assert all(item.debounce is True for item in elements_of_input_and_type_text_url)
    assert all(
        item1.label == item2.display_name
        for item1, item2 in zip(
            elements_of_input_and_type_text_url,
            variable_identifier_input,
            strict=False,
        )
    )


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    INPUT_FIELD_SECTION,
)
def test_build_input_fields_checkbox_components(
    field_list,
    variable,
):
    """Test checkbox fields for variabel identifiers."""
    input_section = build_input_field_section(field_list, "left", variable)
    elements_of_checkbox = [
        element
        for element in input_section.children
        if isinstance(element, ssb.Checkbox)
    ]

    variable_identifier_checkbox = [
        element
        for element in VARIABLES_METADATA
        if isinstance(element, MetadataCheckboxField)
    ]

    assert all(isinstance(item, ssb.Checkbox) for item in elements_of_checkbox)
    assert all(item.disabled is False for item in elements_of_checkbox)
    assert all(item._type == "Checkbox" for item in elements_of_checkbox)  # noqa: SLF001
    assert all(
        item1.label == item2.display_name
        for item1, item2 in zip(
            elements_of_checkbox,
            variable_identifier_checkbox,
            strict=False,
        )
    )


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    INPUT_FIELD_SECTION,
)
def test_build_input_fields_type_date(field_list, variable):
    """Test Input field type 'url'."""
    input_section = build_input_field_section(field_list, "left", variable)

    elements_of_input = [
        element for element in input_section.children if isinstance(element, ssb.Input)
    ]

    elements_of_date = [
        element for element in elements_of_input if element.type == "date"
    ]

    variable_identifier_date = [
        element for element in field_list if isinstance(element, MetadataPeriodField)
    ]

    assert all(
        item1.label == item2.display_name
        for item1, item2 in zip(
            elements_of_date,
            variable_identifier_date,
            strict=False,
        )
    )
    assert all(item.debounce is False for item in elements_of_date)


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    INPUT_FIELD_SECTION,
)
def test_build_input_fields_type_url(field_list, variable):
    input_section = build_input_field_section(field_list, "right", variable)
    variable_identifier_input = [
        element for element in field_list if isinstance(element, MetadataInputField)
    ]
    variable_identifier_url = [
        element for element in variable_identifier_input if element.type == "url"
    ]
    elements_of_input_and_type_url = [
        element
        for element in input_section.children
        if isinstance(element, ssb.Input) and element.type == "url"
    ]
    assert all(item.debounce is True for item in elements_of_input_and_type_url)
    for item1, item2 in zip(
        elements_of_input_and_type_url,
        variable_identifier_url,
        strict=False,
    ):
        assert item1.label == item2.display_name


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("field_list", "variable"),
    INPUT_FIELD_SECTION,
)
def test_build_input_fields_dropdown_components(
    field_list,
    variable,
):
    """Test props for variable identifiers fields."""
    input_section = build_input_field_section(field_list, "right", variable)
    type_dropdown = ssb.Dropdown
    elements_of_dropdown = [
        element
        for element in input_section.children
        if isinstance(element, type_dropdown)
    ]
    variable_identifier_dropdown = [
        element for element in field_list if isinstance(element, MetadataDropdownField)
    ]
    assert all(isinstance(item, type_dropdown) for item in elements_of_dropdown)
    for item in elements_of_dropdown:
        assert item._type == "Dropdown"  # noqa: SLF001
    for item1, item2 in zip(
        elements_of_dropdown,
        variable_identifier_dropdown,
        strict=False,
    ):
        assert item1.header == item2.display_name
    for item1, item2 in zip(
        elements_of_dropdown,
        variable_identifier_dropdown,
        strict=False,
    ):
        assert item1.items == item2.options_getter()



"""Callback functions to do with variables metadata."""

from __future__ import annotations

import datetime
import logging
import urllib.parse
from typing import TYPE_CHECKING

from dapla_metadata.datasets import model

from datadoc_editor import state
from datadoc_editor.constants import DELETE_SELECTED
from datadoc_editor.enums import PseudonymizationAlgorithmsEnum
from datadoc_editor.frontend.callbacks.utils import MetadataInputTypes
from datadoc_editor.frontend.callbacks.utils import PseudonymizationInputTypes
from datadoc_editor.frontend.callbacks.utils import apply_pseudonymization
from datadoc_editor.frontend.callbacks.utils import delete_pseudonymization
from datadoc_editor.frontend.callbacks.utils import find_existing_language_string
from datadoc_editor.frontend.callbacks.utils import map_dropdown_to_pseudo
from datadoc_editor.frontend.callbacks.utils import (
    map_selected_algorithm_to_pseudo_fields,
)
from datadoc_editor.frontend.callbacks.utils import parse_and_validate_dates
from datadoc_editor.frontend.callbacks.utils import (
    parse_and_validate_pseudonymization_time,
)
from datadoc_editor.frontend.callbacks.utils import update_selected_pseudonymization
from datadoc_editor.frontend.components.builders import build_edit_section
from datadoc_editor.frontend.components.builders import build_pseudo_field_section
from datadoc_editor.frontend.components.builders import build_ssb_accordion
from datadoc_editor.frontend.components.builders import build_variables_machine_section
from datadoc_editor.frontend.components.builders import (
    build_variables_pseudonymization_section,
)
from datadoc_editor.frontend.constants import INVALID_DATE_ORDER
from datadoc_editor.frontend.constants import INVALID_VALUE
from datadoc_editor.frontend.constants import PSEUDONYMIZATION
from datadoc_editor.frontend.fields.display_pseudo_variables import (
    PseudoVariableIdentifiers,
)
from datadoc_editor.frontend.fields.display_variables import DISPLAY_VARIABLES
from datadoc_editor.frontend.fields.display_variables import (
    MULTIPLE_LANGUAGE_VARIABLES_METADATA,
)
from datadoc_editor.frontend.fields.display_variables import (
    NON_EDITABLE_VARIABLES_METADATA,
)
from datadoc_editor.frontend.fields.display_variables import VARIABLES_METADATA_LEFT
from datadoc_editor.frontend.fields.display_variables import VARIABLES_METADATA_RIGHT
from datadoc_editor.frontend.fields.display_variables import VariableIdentifiers

if TYPE_CHECKING:
    import dash_bootstrap_components as dbc
    from dapla_metadata.datasets import model


logger = logging.getLogger(__name__)


def populate_variables_workspace(
    variables: list[model.Variable],
    search_query: str,
    dataset_opened_counter: int,
) -> list:
    """Create variable workspace with accordions for variables.

    Allows for filtering which variables are displayed via the search box.
    """
    return [
        build_ssb_accordion(
            variable.short_name or "",
            {
                "type": "variables-accordion",
                "id": f"{variable.short_name}-{dataset_opened_counter}",  # Insert language into the ID to invalidate browser caches
            },
            variable.short_name or "",
            children=[
                build_edit_section(
                    [VARIABLES_METADATA_LEFT, VARIABLES_METADATA_RIGHT],  # type: ignore [list-item]
                    variable,
                ),
                build_variables_machine_section(
                    NON_EDITABLE_VARIABLES_METADATA,
                    "Maskingenerert",
                    variable,
                ),
                build_variables_pseudonymization_section(
                    PSEUDONYMIZATION,
                    variable,
                    map_dropdown_to_pseudo(variable),
                ),
            ],
        )
        for variable in variables
        if search_query in (variable.short_name or "")
    ]


def handle_multi_language_metadata(
    metadata_field: str,
    new_value: MetadataInputTypes | model.LanguageStringType,
    updated_row_id: str,
    language: str,
) -> MetadataInputTypes | model.LanguageStringType:
    """Handle updates to fields which support multiple languages."""
    if new_value is None:
        # This edge case occurs when the user removes the text in an input field
        # We want to ensure we only remove the content for the current language,
        # not create a new blank object!
        return find_existing_language_string(
            state.metadata.variables_lookup[updated_row_id],
            "",
            metadata_field,
            language,
        )

    if isinstance(new_value, str):
        return find_existing_language_string(
            state.metadata.variables_lookup[urllib.parse.unquote(updated_row_id)],
            new_value,
            metadata_field,
            language,
        )

    return new_value


def accept_variable_metadata_input(
    value: MetadataInputTypes,
    variable_short_name: str,
    metadata_field: str,
    language: str | None = None,
) -> str | None:
    """Validate and save the value when variable metadata is updated.

    Returns an error message if an exception was raised, otherwise returns None.
    """
    logger.debug(
        "Updating %s, %s with %s",
        metadata_field,
        variable_short_name,
        value,
    )
    try:
        if (
            metadata_field in MULTIPLE_LANGUAGE_VARIABLES_METADATA
            and language is not None
        ):
            new_value = handle_multi_language_metadata(
                metadata_field,
                value,
                variable_short_name,
                language,
            )
        elif value == "":
            # Allow clearing non-multiple-language text fields
            new_value = None
        else:
            new_value = value

        # Write the value to the variables structure
        setattr(
            state.metadata.variables_lookup[urllib.parse.unquote(variable_short_name)],
            metadata_field,
            new_value,
        )
    except ValueError:
        logger.exception(
            "Validation failed for %s, %s, %s:",
            metadata_field,
            variable_short_name,
            value,
        )
        return INVALID_VALUE
    else:
        if value == "":
            value = None
        logger.info(
            "Updated %s: %s with value '%s'",
            variable_short_name,
            metadata_field,
            value,
        )
        return None


def accept_pseudo_variable_metadata_input(
    value: PseudonymizationInputTypes,
    variable_short_name: str,
    metadata_field: str,
) -> str | None:
    """Validate and save the value when a pseudo variable metadata is updated.

    If metadata field is 'pseudonymization_time' date is parsed and validated, else value
    is stripped from all whitespace.

    Returns an error message if an exception was raised, otherwise returns None.
    """
    logger.debug(
        "Updating %s, %s with %s",
        metadata_field,
        variable_short_name,
        value,
    )
    variable_pseudonymization = state.metadata.variables_lookup[
        urllib.parse.unquote(variable_short_name)
    ].pseudonymization
    try:
        parsed_value: str | datetime.datetime | None
        if (
            metadata_field == PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME
            and isinstance(value, (datetime.datetime, str))
        ):
            parsed_value = parse_and_validate_pseudonymization_time(value)
        elif isinstance(value, str):
            parsed_value = value.strip() or None
        else:
            parsed_value = value or None
        setattr(
            variable_pseudonymization,
            metadata_field,
            parsed_value,
        )
    except ValueError:
        logger.exception(
            "Validation failed for %s, %s, %s:",
            metadata_field,
            variable_short_name,
            value,
        )
        return INVALID_VALUE
    else:
        logger.info(
            "Updated %s: %s with value '%s'",
            variable_short_name,
            metadata_field,
            value,
        )
        return None


def accept_variable_metadata_date_input(
    variable_identifier: VariableIdentifiers,
    variable_short_name: str,
    contains_data_from: str,
    contains_data_until: str,
) -> tuple[bool, str, bool, str]:
    """Validate and save date range inputs."""
    message = ""

    try:
        (
            parsed_contains_data_from,
            parsed_contains_data_until,
        ) = parse_and_validate_dates(
            str(
                contains_data_from
                or state.metadata.variables_lookup[
                    variable_short_name
                ].contains_data_from,
            ),
            str(
                contains_data_until
                or state.metadata.variables_lookup[
                    variable_short_name
                ].contains_data_until,
            ),
        )

        # Save both values to the model if they pass validation.
        state.metadata.variables_lookup[
            variable_short_name
        ].contains_data_from = parsed_contains_data_from
        state.metadata.variables_lookup[
            variable_short_name
        ].contains_data_until = parsed_contains_data_until
    except ValueError as e:
        logger.exception(
            "Validation failed for %s, %s, %s: %s, %s: %s",
            variable_identifier,
            variable_short_name,
            "contains_data_from",
            contains_data_from,
            "contains_data_until",
            contains_data_until,
        )
        message = str(e)
    else:
        logger.debug(
            "Successfully updated %s, %s, %s: %s, %s: %s",
            variable_identifier,
            variable_short_name,
            "contains_data_from",
            contains_data_from,
            "contains_data_until",
            contains_data_until,
        )

    no_error = (False, "")
    if not message:
        # No error to display.
        return no_error + no_error

    error = (
        True,
        INVALID_DATE_ORDER.format(
            contains_data_from_display_name=DISPLAY_VARIABLES[
                VariableIdentifiers.CONTAINS_DATA_FROM
            ].display_name,
            contains_data_until_display_name=DISPLAY_VARIABLES[
                VariableIdentifiers.CONTAINS_DATA_UNTIL
            ].display_name,
        ),
    )
    return (
        error + no_error
        if variable_identifier == VariableIdentifiers.CONTAINS_DATA_FROM
        else no_error + error
    )


def variable_identifier(
    dataset_identifier: str,
) -> str | None:
    """Pair corresponding identifiers."""
    metadata_identifiers = {
        "temporality_type": VariableIdentifiers.TEMPORALITY_TYPE,
        "data_source": VariableIdentifiers.DATA_SOURCE,
        "contains_data_from": VariableIdentifiers.CONTAINS_DATA_FROM,
        "contains_data_until": VariableIdentifiers.CONTAINS_DATA_UNTIL,
    }
    return metadata_identifiers.get(dataset_identifier)


def variable_identifier_multilanguage(
    dataset_identifier: str,
) -> str | None:
    """Pair corresponding identifiers for multilanguage fields."""
    metadata_identifiers = {
        "population_description": VariableIdentifiers.POPULATION_DESCRIPTION,
    }
    return metadata_identifiers.get(dataset_identifier)


def set_variables_values_inherit_dataset_values(
    value: MetadataInputTypes | model.LanguageStringType,
    metadata_identifier: str,
) -> None:
    """Set variable value based on dataset value."""
    variable = variable_identifier(metadata_identifier)
    if value is not None and variable is not None:
        for val in state.metadata.variables:
            setattr(
                state.metadata.variables_lookup[val.short_name],
                variable,
                value,
            )


def set_variables_value_multilanguage_inherit_dataset_values(
    value: MetadataInputTypes | model.LanguageStringType,
    metadata_identifier: str,
    language: str,
) -> None:
    """Set variable multilanguage value based on dataset value."""
    variable = variable_identifier_multilanguage(metadata_identifier)
    if value is not None and variable is not None:
        for val in state.metadata.variables:
            update_value = handle_multi_language_metadata(
                variable,
                value,
                val.short_name,
                language,
            )
            setattr(
                state.metadata.variables_lookup[val.short_name],
                variable,
                update_value,
            )


def set_variables_values_inherit_dataset_derived_date_values() -> None:
    """Set variable date values if variables date values are not set.

    Covers the case for inherit dataset date values where dates are derived from dataset path
    and must be set on file opening.
    """
    for val in state.metadata.variables:
        if state.metadata.variables_lookup[val.short_name].contains_data_from is None:
            setattr(
                state.metadata.variables_lookup[val.short_name],
                VariableIdentifiers.CONTAINS_DATA_FROM,
                state.metadata.dataset.contains_data_from,
            )
        if state.metadata.variables_lookup[val.short_name].contains_data_until is None:
            setattr(
                state.metadata.variables_lookup[val.short_name],
                VariableIdentifiers.CONTAINS_DATA_UNTIL,
                state.metadata.dataset.contains_data_until,
            )


def populate_pseudo_workspace(
    variable: model.Variable,
    selected_algorithm: PseudonymizationAlgorithmsEnum | None,
) -> dbc.Form:
    """Build pseudonymization workspace for a variable.

    Infers or applies variable pseudonymization.
    Builds pseudonymization workspace dynamically based on selected pseudo algorithm.

    Args:
        variable (model.Variable):
            The variable to build the pseudonymization workspace for.
        selected_algorithm (PseudonymizationAlgorithmsEnum | str | None):
            The pseudonymization algorithm selected by the user.
            If None and the variable already has pseudonymization, the algorithm
            is inferred from the existing state.

    Returns:
        dbc.Form | list:
            Pseudonymization fields based on selection if pseudonymization; otherwise, an empty list.
    """
    if not selected_algorithm and variable.pseudonymization:
        selected_algorithm = map_dropdown_to_pseudo(variable)

    if selected_algorithm and not variable.pseudonymization:
        apply_pseudonymization(variable, selected_algorithm, None)
        logger.info("Added pseudonymization for %s", variable.short_name)

    if variable.pseudonymization is None:
        logger.info(
            "No pseudonymization for %s, returning empty list", variable.short_name
        )
        return []

    return build_pseudo_field_section(
        map_selected_algorithm_to_pseudo_fields(selected_algorithm),
        "left",
        variable=variable,
        pseudonymization=variable.pseudonymization,
        field_id="pseudo",
    )


def mutate_variable_pseudonymization(
    variable: model.Variable,
    selected_algorithm: PseudonymizationAlgorithmsEnum | str | None,
) -> None:
    """Updates or delete variable pseudonymization.

    Depending on the selected algorithm, this function will update the existing
    pseudonymization or delete it.

    Args:
        variable (model.Variable):
            The variable whose pseudonymization should be mutated.
        selected_algorithm (PseudonymizationAlgorithmsEnum | str | None):
            The pseudonymization algorithm selected by the user. If equal to
            "delete_selected", the pseudonymization is removed. If an algorithm is
            provided and differs from the inferred algorithm, the pseudonymization
            is updated accordingly.
    """
    if selected_algorithm == DELETE_SELECTED and variable.pseudonymization:
        delete_pseudonymization(variable)
        return

    if (
        isinstance(selected_algorithm, PseudonymizationAlgorithmsEnum)
        and variable.pseudonymization
    ):
        inferred_algorithm = map_dropdown_to_pseudo(variable)
        if inferred_algorithm and inferred_algorithm != selected_algorithm:
            update_selected_pseudonymization(
                variable, inferred_algorithm, selected_algorithm
            )
        return



import pytest

from datadoc_editor import state
from datadoc_editor.frontend.fields.display_base import DROPDOWN_DESELECT_OPTION
from datadoc_editor.frontend.fields.display_dataset import (
    get_statistical_subject_options,
)
from tests.conftest import STATISTICAL_SUBJECT_STRUCTURE_DIR
from tests.utils import TEST_RESOURCES_DIRECTORY


@pytest.mark.parametrize(
    ("subject_xml_file_path", "expected"),
    [
        (
            TEST_RESOURCES_DIRECTORY
            / STATISTICAL_SUBJECT_STRUCTURE_DIR
            / "extract_secondary_subject.xml",
            [
                {"title": DROPDOWN_DESELECT_OPTION, "id": ""},
                {"title": "aa norwegian - aa00 norwegian", "id": "aa00"},
                {"title": "aa norwegian - aa01 norwegian", "id": "aa01"},
                {"title": "ab norwegian - ab00 norwegian", "id": "ab00"},
                {"title": "ab norwegian - ab01 norwegian", "id": "ab01"},
            ],
        ),
        (
            TEST_RESOURCES_DIRECTORY
            / STATISTICAL_SUBJECT_STRUCTURE_DIR
            / "missing_language.xml",
            [
                {"title": DROPDOWN_DESELECT_OPTION, "id": ""},
                {"title": " - aa00 norwegian", "id": "aa00"},
                {"title": " - aa01 norwegian", "id": "aa01"},
                {"title": " - ab00 norwegian", "id": "ab00"},
                {"title": " - ", "id": "ab01"},
            ],
        ),
    ],
)
def test_get_statistical_subject_options(
    subject_mapping_fake_statistical_structure,
    expected,
):
    state.statistic_subject_mapping = subject_mapping_fake_statistical_structure
    state.statistic_subject_mapping.wait_for_external_result()
    assert get_statistical_subject_options() == expected



"""Test function build_ssb_accordion."""

import pytest
import ssb_dash_components as ssb  # type: ignore[import-untyped]

from datadoc_editor.frontend.components.builders import build_ssb_accordion
from datadoc_editor.frontend.fields.display_variables import (
    OBLIGATORY_VARIABLES_METADATA,
)
from datadoc_editor.frontend.fields.display_variables import OPTIONAL_VARIABLES_METADATA

ACCORDION_TYPE = "variables-accordion"


@pytest.mark.parametrize(
    ("name", "dict_id", "field_list"),
    [
        (
            "pers_id",
            {"type": ACCORDION_TYPE, "id": "pers_id"},
            [],
        ),
        (
            "sykepenger",
            {"type": ACCORDION_TYPE, "id": "sykepenger"},
            OPTIONAL_VARIABLES_METADATA,
        ),
        (
            "ber_bruttoformue",
            {"type": ACCORDION_TYPE, "id": "ber_bruttoformue"},
            OBLIGATORY_VARIABLES_METADATA,
        ),
        (
            "hoveddiagnose",
            {"type": ACCORDION_TYPE, "id": "hoveddiagnose"},
            [],
        ),
        (
            "",
            {"type": ACCORDION_TYPE, "id": ""},
            OBLIGATORY_VARIABLES_METADATA,
        ),
    ],
)
def test_build_ssb_accordion(name, dict_id, field_list):
    accordion = build_ssb_accordion(name, dict_id, name, field_list)
    assert accordion.id["type"] == ACCORDION_TYPE
    assert accordion.id["id"] == accordion.header
    assert isinstance(accordion, ssb.Accordion)



"""Functionality for displaying dataset metadata."""

from __future__ import annotations

import functools
import logging
from enum import Enum

from dapla_metadata.datasets import enums

from datadoc_editor import state
from datadoc_editor.enums import Assessment
from datadoc_editor.enums import DataSetState
from datadoc_editor.enums import DataSetStatus
from datadoc_editor.enums import UseRestrictionType
from datadoc_editor.frontend.fields.display_base import DATASET_METADATA_DATE_INPUT
from datadoc_editor.frontend.fields.display_base import (
    DATASET_METADATA_MULTIDROPDOWN_INPUT,
)
from datadoc_editor.frontend.fields.display_base import (
    DATASET_METADATA_MULTILANGUAGE_INPUT,
)
from datadoc_editor.frontend.fields.display_base import DROPDOWN_DESELECT_OPTION
from datadoc_editor.frontend.fields.display_base import FieldTypes
from datadoc_editor.frontend.fields.display_base import MetadataDateField
from datadoc_editor.frontend.fields.display_base import MetadataDropdownField
from datadoc_editor.frontend.fields.display_base import MetadataInputField
from datadoc_editor.frontend.fields.display_base import MetadataMultiDropdownField
from datadoc_editor.frontend.fields.display_base import MetadataMultiLanguageField
from datadoc_editor.frontend.fields.display_base import MetadataPeriodField
from datadoc_editor.frontend.fields.display_base import get_comma_separated_string
from datadoc_editor.frontend.fields.display_base import get_enum_options

logger = logging.getLogger(__name__)


def get_statistical_subject_options() -> list[dict[str, str]]:
    """Generate the list of options for statistical subject."""
    dropdown_options = [
        {
            "title": f"{primary.get_title(enums.SupportedLanguages.NORSK_BOKMÅL)} - {secondary.get_title(enums.SupportedLanguages.NORSK_BOKMÅL)}",
            "id": secondary.subject_code,
        }
        for primary in state.statistic_subject_mapping.primary_subjects
        for secondary in primary.secondary_subjects
    ]
    dropdown_options.insert(0, {"title": DROPDOWN_DESELECT_OPTION, "id": ""})
    return dropdown_options


def get_owner_options() -> list[dict[str, str]]:
    """Collect the owner options."""
    dropdown_options = [
        {
            "title": f"{option.code} - {option.get_title(enums.SupportedLanguages.NORSK_BOKMÅL)}",
            "id": option.code,
        }
        for option in state.organisational_units.classifications
    ]
    dropdown_options.insert(0, {"title": DROPDOWN_DESELECT_OPTION, "id": ""})
    return dropdown_options


class DatasetIdentifiers(str, Enum):
    """As defined here: https://statistics-norway.atlassian.net/l/c/aoSfEWJU."""

    SHORT_NAME = "short_name"
    ASSESSMENT = "assessment"
    DATASET_STATUS = "dataset_status"
    DATASET_STATE = "dataset_state"
    NAME = "name"
    DESCRIPTION = "description"
    POPULATION_DESCRIPTION = "population_description"
    VERSION = "version"
    VERSION_DESCRIPTION = "version_description"
    SUBJECT_FIELD = "subject_field"
    KEYWORD = "keyword"
    SPATIAL_COVERAGE_DESCRIPTION = "spatial_coverage_description"
    USE_RESTRICTIONS = "use_restrictions"
    ID = "id"
    OWNER = "owner"
    FILE_PATH = "file_path"
    METADATA_CREATED_DATE = "metadata_created_date"
    METADATA_CREATED_BY = "metadata_created_by"
    METADATA_LAST_UPDATED_DATE = "metadata_last_updated_date"
    METADATA_LAST_UPDATED_BY = "metadata_last_updated_by"
    CONTAINS_DATA_FROM = "contains_data_from"
    CONTAINS_DATA_UNTIL = "contains_data_until"


DISPLAY_DATASET: dict[
    DatasetIdentifiers,
    FieldTypes,
] = {
    DatasetIdentifiers.NAME: MetadataMultiLanguageField(
        identifier=DatasetIdentifiers.NAME.value,
        display_name="Navn",
        description="Oppgi navn på datasettet. Navnet skal være forståelig for mennesker slik at det er søkbart.",
        obligatory=True,
        id_type=DATASET_METADATA_MULTILANGUAGE_INPUT,
    ),
    DatasetIdentifiers.DESCRIPTION: MetadataMultiLanguageField(
        identifier=DatasetIdentifiers.DESCRIPTION.value,
        display_name="Beskrivelse",
        description="Beskrivelse av datasettet",
        obligatory=True,
        id_type=DATASET_METADATA_MULTILANGUAGE_INPUT,
    ),
    DatasetIdentifiers.ASSESSMENT: MetadataDropdownField(
        identifier=DatasetIdentifiers.ASSESSMENT.value,
        display_name="Verdivurdering",
        description="Verdivurderingen utledes fra datatilstanden og kan ha verdiene: sensitiv (kildedata), skjermet (inndata, klargjorte data og statistikk) og åpen (utdata).",
        obligatory=True,
        options_getter=functools.partial(
            get_enum_options,
            Assessment,
        ),
    ),
    DatasetIdentifiers.POPULATION_DESCRIPTION: MetadataMultiLanguageField(
        identifier=DatasetIdentifiers.POPULATION_DESCRIPTION.value,
        display_name="Populasjon",
        description='Oppgi populasjonen datasettet dekker. Beskrivelsen skal inkludere enhetstype, geografisk dekningsområde og tidsperiode, f.eks.  "Personer bosatt  i Norge 1990-2020"',
        obligatory=True,
        id_type=DATASET_METADATA_MULTILANGUAGE_INPUT,
    ),
    DatasetIdentifiers.USE_RESTRICTIONS: MetadataMultiDropdownField(
        identifier=DatasetIdentifiers.USE_RESTRICTIONS.value,
        display_name="Bruksrestriksjon",
        description="Velg hvilken bruksrestriksjon som gjelder.",
        options_getter=functools.partial(get_enum_options, UseRestrictionType),
        type_display_name="Bruksrestriksjon",
        type_description="Oppgi om det er knyttet noen bruksrestriksjoner til datasettet, f.eks. krav om sletting/anonymisering.",
        date_display_name="Dato for restriksjon",
        date_description='Oppgi ev. "tiltaksdato" for bruksrestriksjoner, f.eks. frist for sletting/anonymisering. Noen bruksrestriksjoner vil ikke ha en slik dato, f.eks. vil en behandlingsbegrensning normalt være permanent/tidsuavhengig.',
        id_type=DATASET_METADATA_MULTIDROPDOWN_INPUT,
    ),
    DatasetIdentifiers.DATASET_STATE: MetadataDropdownField(
        identifier=DatasetIdentifiers.DATASET_STATE.value,
        display_name="Datatilstand",
        description="Datasettets datatilstand der en av de følgende er mulige: kildedata, inndata, klargjorte data, statistikk og utdata.",
        obligatory=True,
        options_getter=functools.partial(
            get_enum_options,
            DataSetState,
        ),
    ),
    DatasetIdentifiers.DATASET_STATUS: MetadataDropdownField(
        identifier=DatasetIdentifiers.DATASET_STATUS.value,
        display_name="Status",
        description="Oppgi om metadataene er under arbeid (utkast), kan deles internt (intern), kan deles eksternt(ekstern) eller er avsluttet/erstattet (utgått). Det kan være restriksjoner knyttet til deling både internt og eksternt.",
        options_getter=functools.partial(
            get_enum_options,
            DataSetStatus,
        ),
        obligatory=True,
    ),
    DatasetIdentifiers.CONTAINS_DATA_FROM: MetadataPeriodField(
        identifier=DatasetIdentifiers.CONTAINS_DATA_FROM.value,
        display_name="Inneholder data f.o.m.",
        description="Oppgi hvilken dato datasettet inneholder data f.o.m. ÅÅÅÅ-MM-DD",
        obligatory=True,
        editable=True,
        id_type=DATASET_METADATA_DATE_INPUT,
    ),
    DatasetIdentifiers.CONTAINS_DATA_UNTIL: MetadataPeriodField(
        identifier=DatasetIdentifiers.CONTAINS_DATA_UNTIL.value,
        display_name="Inneholder data t.o.m.",
        description="Oppgi hvilken dato datasettet inneholder data t.o.m. ÅÅÅÅ-MM-DD",
        obligatory=True,
        editable=True,
        id_type=DATASET_METADATA_DATE_INPUT,
    ),
    DatasetIdentifiers.SUBJECT_FIELD: MetadataDropdownField(
        identifier=DatasetIdentifiers.SUBJECT_FIELD.value,
        display_name="Statistikkområde",
        description="Oppgi det primære statistikkområdet som datasettet tilhører.",
        obligatory=True,
        options_getter=get_statistical_subject_options,
    ),
    DatasetIdentifiers.KEYWORD: MetadataInputField(
        identifier=DatasetIdentifiers.KEYWORD.value,
        display_name="Nøkkelord",
        description="Her kan en oppgi nøkkelord som beskriver datasettet, og som kan brukes i søk. Nøkkelordene må legges inn som en kommaseparert streng. F.eks. befolkning, skatt, arbeidsledighet",
        value_getter=get_comma_separated_string,
    ),
    DatasetIdentifiers.VERSION: MetadataInputField(
        identifier=DatasetIdentifiers.VERSION.value,
        display_name="Versjon",
        description="Oppgi hvilken versjon av datasettet dette er (versjonering av datasett er beskrevet i Dapla-manualen).",
        obligatory=True,
        type="number",
    ),
    DatasetIdentifiers.VERSION_DESCRIPTION: MetadataMultiLanguageField(
        identifier=DatasetIdentifiers.VERSION_DESCRIPTION.value,
        display_name="Versjonsbeskrivelse",
        description="Beskriv kort årsaken til at denne versjonen av datasettet ble laget.",
        obligatory=True,
        id_type=DATASET_METADATA_MULTILANGUAGE_INPUT,
    ),
    DatasetIdentifiers.SPATIAL_COVERAGE_DESCRIPTION: MetadataMultiLanguageField(
        identifier=DatasetIdentifiers.SPATIAL_COVERAGE_DESCRIPTION.value,
        display_name="Geografisk dekningsområde",
        description="Oppgi datasettets geografiske dekningsområde, f.eks. Norge.",
        obligatory=True,
        id_type=DATASET_METADATA_MULTILANGUAGE_INPUT,
    ),
    DatasetIdentifiers.SHORT_NAME: MetadataInputField(
        identifier=DatasetIdentifiers.SHORT_NAME.value,
        display_name="Kortnavn",
        description="Datasettets tekniske kortnavn (uten versjonsnummer og filendelse).",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.ID: MetadataInputField(
        identifier=DatasetIdentifiers.ID.value,
        display_name="ID",
        description="Den unike SSB-identifikatoren for datasettet.",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.FILE_PATH: MetadataInputField(
        identifier=DatasetIdentifiers.FILE_PATH.value,
        display_name="Filsti",
        description="Filstien inneholder datasettets kortnavn og stien til stedet der det er lagret.",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_CREATED_DATE: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_CREATED_DATE.value,
        display_name="Dato opprettet",
        description="Datoen metadataene for datasettet ble opprettet",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_CREATED_BY: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_CREATED_BY.value,
        display_name="Opprettet av",
        description=" Navnet på personen som opprettet metadataene",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_LAST_UPDATED_DATE: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_LAST_UPDATED_DATE.value,
        display_name="Dato oppdatert",
        description="Datoen metadataene om datasettet sist ble oppdatert",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.METADATA_LAST_UPDATED_BY: MetadataInputField(
        identifier=DatasetIdentifiers.METADATA_LAST_UPDATED_BY.value,
        display_name="Oppdatert av",
        description="Navnet på personen som sist oppdaterte metadataene.",
        obligatory=True,
        editable=False,
    ),
    DatasetIdentifiers.OWNER: MetadataInputField(
        identifier=DatasetIdentifiers.OWNER.value,
        display_name="Eier",
        description="Navnet på teamet som eier datasettet",
        obligatory=True,
        editable=False,
    ),
}

MULTIPLE_LANGUAGE_DATASET_IDENTIFIERS = [
    m.identifier
    for m in DISPLAY_DATASET.values()
    if isinstance(m, MetadataMultiLanguageField)
]

MULTIPLE_DROPDOWN_DATASET_IDENTIFIERS = [
    m.identifier
    for m in DISPLAY_DATASET.values()
    if isinstance(m, MetadataMultiDropdownField)
]

EDITABLE_DATASET_METADATA_LEFT = [
    m
    for m in DISPLAY_DATASET.values()
    if m.editable and isinstance(m, MetadataMultiLanguageField)
]

EDITABLE_DATASET_METADATA_LEFT.insert(3, DISPLAY_DATASET[DatasetIdentifiers.VERSION])


EDITABLE_DATASET_METADATA_RIGHT = [
    m
    for m in DISPLAY_DATASET.values()
    if m.editable and m not in EDITABLE_DATASET_METADATA_LEFT
]


NON_EDITABLE_DATASET_METADATA = [m for m in DISPLAY_DATASET.values() if not m.editable]

# The order of this list MUST match the order of display components, as defined in DatasetTab.py
DISPLAYED_DATASET_METADATA: list[FieldTypes] = (
    EDITABLE_DATASET_METADATA_LEFT
    + EDITABLE_DATASET_METADATA_RIGHT
    + NON_EDITABLE_DATASET_METADATA
)

DROPDOWN_DATASET_METADATA: list[MetadataDropdownField] = [
    m for m in DISPLAYED_DATASET_METADATA if isinstance(m, MetadataDropdownField)
]

DROPDOWN_DATASET_METADATA_IDENTIFIERS: list[str] = [
    m.identifier for m in DROPDOWN_DATASET_METADATA
]

TIMEZONE_AWARE_METADATA_IDENTIFIERS = [
    m.identifier for m in DISPLAYED_DATASET_METADATA if isinstance(m, MetadataDateField)
]

OBLIGATORY_DATASET_METADATA_IDENTIFIERS_AND_DISPLAY_NAME: list[tuple] = [
    (m.identifier, m.display_name)
    for m in DISPLAY_DATASET.values()
    if m.obligatory and m.editable
]



"""Components and layout which are not inside a tab."""

from __future__ import annotations

import ssb_dash_components as ssb
from dash import html

from datadoc_editor.frontend.callbacks.utils import get_dataset_path
from datadoc_editor.utils import get_app_version

header = ssb.Header(
    [
        ssb.Title("Datadoc Editor", size=1, id="main-title", className="main-title"),
        ssb.Link(
            "Dokumentasjon",
            href="https://manual.dapla.ssb.no/statistikkere/datadoc.html",
            isExternal=True,
        ),
    ],
    className="datadoc-header",
)


def build_footer_control_bar() -> html.Aside:
    """Build footer control bar which resides below all the content."""
    return html.Aside(
        children=[
            html.P(
                f"v{get_app_version()}",
                className="small",
            ),
        ],
        className="language-footer",
    )


def build_controls_bar() -> html.Section:
    """Build the Controls Bar.

    This contains:
    - A text input to specify the path to a dataset
    - A button to open a dataset
    - A button to save metadata to disk
    """
    return html.Section(
        [
            ssb.Input(
                label="Filsti",
                value=get_dataset_path(),
                className="file-path-input",
                id="dataset-path-input",
            ),
            html.Section(
                [
                    ssb.Button(
                        children=["Åpne fil"],
                        id="open-button",
                        className="file-open-button",
                    ),
                    ssb.Button(
                        children=["Lagre metadata"],
                        id="save-button",
                        className="file-save-button",
                    ),
                ],
                className="button-section",
            ),
        ],
        className="control-bar-section",
    )



"""Tests for the variables callbacks module."""

from __future__ import annotations

import datetime
import warnings
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from uuid import UUID

import arrow
import dash_bootstrap_components as dbc
import pytest
from dapla_metadata.datasets import ObligatoryVariableWarning
from dapla_metadata.datasets import model
from pydantic import AnyUrl

from datadoc_editor import constants
from datadoc_editor import enums
from datadoc_editor import state
from datadoc_editor.frontend.callbacks.utils import update_selected_pseudonymization
from datadoc_editor.frontend.callbacks.utils import variables_control
from datadoc_editor.frontend.callbacks.variables import (
    accept_pseudo_variable_metadata_input,
)
from datadoc_editor.frontend.callbacks.variables import (
    accept_variable_metadata_date_input,
)
from datadoc_editor.frontend.callbacks.variables import accept_variable_metadata_input
from datadoc_editor.frontend.callbacks.variables import mutate_variable_pseudonymization
from datadoc_editor.frontend.callbacks.variables import populate_pseudo_workspace
from datadoc_editor.frontend.callbacks.variables import populate_variables_workspace
from datadoc_editor.frontend.callbacks.variables import (
    set_variables_value_multilanguage_inherit_dataset_values,
)
from datadoc_editor.frontend.callbacks.variables import (
    set_variables_values_inherit_dataset_derived_date_values,
)
from datadoc_editor.frontend.callbacks.variables import (
    set_variables_values_inherit_dataset_values,
)
from datadoc_editor.frontend.constants import INVALID_DATE_ORDER
from datadoc_editor.frontend.constants import INVALID_VALUE
from datadoc_editor.frontend.fields.display_base import get_metadata_and_stringify
from datadoc_editor.frontend.fields.display_base import get_standard_metadata
from datadoc_editor.frontend.fields.display_dataset import DatasetIdentifiers
from datadoc_editor.frontend.fields.display_pseudo_variables import (
    PseudoVariableIdentifiers,
)
from datadoc_editor.frontend.fields.display_variables import DISPLAY_VARIABLES
from datadoc_editor.frontend.fields.display_variables import VariableIdentifiers

if TYPE_CHECKING:
    from dapla_metadata.datasets import Datadoc

    from datadoc_editor.frontend.callbacks.utils import MetadataInputTypes


@pytest.fixture
def n_clicks_1():
    return 1


@pytest.mark.parametrize(
    ("metadata_field", "value", "expected_model_value"),
    [
        (
            VariableIdentifiers.NAME,
            "Variable name",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Variable name",
                    ),
                ],
            ),
        ),
        (
            VariableIdentifiers.DATA_TYPE,
            enums.DataType.STRING,
            enums.DataType.STRING.value,
        ),
        (
            VariableIdentifiers.VARIABLE_ROLE,
            enums.VariableRole.MEASURE,
            enums.VariableRole.MEASURE.value,
        ),
        (
            VariableIdentifiers.DEFINITION_URI,
            "https://www.example.com",
            AnyUrl("https://www.example.com"),
        ),
        (
            VariableIdentifiers.IS_PERSONAL_DATA,
            False,
            False,
        ),
        (
            VariableIdentifiers.UNIT_TYPE,
            "17",
            "17",
        ),
        (
            VariableIdentifiers.DATA_SOURCE,
            "Atlantis",
            "Atlantis",
        ),
        (
            VariableIdentifiers.POPULATION_DESCRIPTION,
            "Population description",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Population description",
                    ),
                ],
            ),
        ),
        (
            VariableIdentifiers.COMMENT,
            "Comment",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Comment",
                    ),
                ],
            ),
        ),
        (
            VariableIdentifiers.TEMPORALITY_TYPE,
            enums.TemporalityTypeType.ACCUMULATED,
            enums.TemporalityTypeType.ACCUMULATED.value,
        ),
        (
            VariableIdentifiers.MEASUREMENT_UNIT,
            "Kilograms",
            "Kilograms",
        ),
        (
            VariableIdentifiers.FORMAT,
            "Regex",
            "Regex",
        ),
        (
            VariableIdentifiers.CLASSIFICATION_URI,
            "https://www.example.com",
            AnyUrl("https://www.example.com"),
        ),
        (
            VariableIdentifiers.INVALID_VALUE_DESCRIPTION,
            "Invalid value",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Invalid value",
                    ),
                ],
            ),
        ),
        (
            VariableIdentifiers.IDENTIFIER,
            "2f72477a-f051-43ee-bf8b-0d8f47b5e0a7",
            UUID("2f72477a-f051-43ee-bf8b-0d8f47b5e0a7"),
        ),
    ],
)
def test_accept_variable_metadata_input_valid(
    metadata: Datadoc,
    metadata_field: VariableIdentifiers,
    value: MetadataInputTypes,
    expected_model_value: Any,  # noqa: ANN401
):
    state.metadata = metadata
    assert (
        accept_variable_metadata_input(
            value,
            metadata.variables[0].short_name,
            metadata_field=metadata_field.value,
            language="nb",
        )
        is None
    )
    assert (
        getattr(state.metadata.variables[0], metadata_field.value)
        == expected_model_value
    )


def test_accept_variable_metadata_input_invalid(
    metadata: Datadoc,
):
    state.metadata = metadata
    message = accept_variable_metadata_input(
        "not a url",
        metadata.variables[0].short_name,
        metadata_field=VariableIdentifiers.DEFINITION_URI.value,
    )
    assert message is not None
    assert message == INVALID_VALUE


@pytest.mark.parametrize(
    (
        "variable_identifier",
        "contains_data_from",
        "contains_data_until",
        "expected_result",
    ),
    [
        (
            VariableIdentifiers.CONTAINS_DATA_FROM.value,
            "1950-01-01",
            "2020-01-01",
            (False, "", False, ""),
        ),
        (
            VariableIdentifiers.CONTAINS_DATA_FROM.value,
            "2020-01-01",
            "1950-01-01",
            (
                True,
                INVALID_DATE_ORDER.format(
                    contains_data_from_display_name=DISPLAY_VARIABLES[
                        VariableIdentifiers.CONTAINS_DATA_FROM
                    ].display_name,
                    contains_data_until_display_name=DISPLAY_VARIABLES[
                        VariableIdentifiers.CONTAINS_DATA_UNTIL
                    ].display_name,
                ),
                False,
                "",
            ),
        ),
        (
            VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            "1950-01-01",
            "2020-01-01",
            (False, "", False, ""),
        ),
        (
            VariableIdentifiers.CONTAINS_DATA_UNTIL.value,
            "2020-01-01",
            "1950-01-01",
            (
                False,
                "",
                True,
                INVALID_DATE_ORDER.format(
                    contains_data_from_display_name=DISPLAY_VARIABLES[
                        VariableIdentifiers.CONTAINS_DATA_FROM
                    ].display_name,
                    contains_data_until_display_name=DISPLAY_VARIABLES[
                        VariableIdentifiers.CONTAINS_DATA_UNTIL
                    ].display_name,
                ),
            ),
        ),
    ],
)
def test_accept_variable_metadata_date_input(
    variable_identifier,
    contains_data_from: str,
    contains_data_until: str,
    expected_result: tuple[bool, str, bool, str],
    metadata: Datadoc,
):
    state.metadata = metadata
    chosen_short_name = metadata.variables[0].short_name
    preset_identifier = (
        VariableIdentifiers.CONTAINS_DATA_UNTIL.value
        if variable_identifier == VariableIdentifiers.CONTAINS_DATA_FROM.value
        else VariableIdentifiers.CONTAINS_DATA_FROM.value
    )
    preset_value = (
        contains_data_until
        if variable_identifier == VariableIdentifiers.CONTAINS_DATA_FROM.value
        else contains_data_from
    )
    setattr(
        state.metadata.variables_lookup[chosen_short_name],
        preset_identifier,
        arrow.get(preset_value).date(),
    )
    assert (
        accept_variable_metadata_date_input(
            VariableIdentifiers(variable_identifier),
            chosen_short_name,
            contains_data_from,
            contains_data_until,
        )
        == expected_result
    )
    if not expected_result[0]:
        assert (
            metadata.variables[0].contains_data_from
            == arrow.get(
                contains_data_from,
            ).date()
        )
    if not expected_result[2]:
        assert (
            metadata.variables[0].contains_data_until
            == arrow.get(
                contains_data_until,
            ).date()
        )


@pytest.mark.usefixtures("_code_list_fake_classifications")
@pytest.mark.parametrize(
    ("search_query", "expected_length"),
    [
        (
            "",
            8,
        ),
        (
            "a",
            4,
        ),
        (
            "pers_id",
            1,
        ),
    ],
)
def test_populate_variables_workspace_filter_variables(
    search_query: str, expected_length: int, metadata: Datadoc
):
    assert (
        len(
            populate_variables_workspace(metadata.variables, search_query, 0),
        )
        == expected_length
    )


@pytest.mark.parametrize(
    (
        "dataset_value",
        "dataset_identifier",
        "variable_identifier",
    ),
    [
        (
            "2009-01-02",
            DatasetIdentifiers.CONTAINS_DATA_FROM,
            VariableIdentifiers.CONTAINS_DATA_FROM,
        ),
        (
            "2021-08-10",
            DatasetIdentifiers.CONTAINS_DATA_UNTIL,
            VariableIdentifiers.CONTAINS_DATA_UNTIL,
        ),
    ],
)
def test_variables_values_inherit_dataset_values(
    dataset_value,
    dataset_identifier,
    variable_identifier,
    metadata: Datadoc,
):
    state.metadata = metadata
    setattr(
        state.metadata.dataset,
        dataset_identifier,
        dataset_value,
    )
    set_variables_values_inherit_dataset_values(
        dataset_value,
        dataset_identifier,
    )
    for val in state.metadata.variables:
        assert dataset_value == get_metadata_and_stringify(
            metadata.variables_lookup[val.short_name],
            variable_identifier.value,
        )


@pytest.mark.parametrize(
    (
        "dataset_value",
        "dataset_identifier",
        "variable_identifier",
        "update_value",
    ),
    [
        (
            "2009-01-02",
            DatasetIdentifiers.CONTAINS_DATA_FROM,
            VariableIdentifiers.CONTAINS_DATA_FROM,
            "1998-03-11",
        ),
        (
            "1988-11-03",
            DatasetIdentifiers.CONTAINS_DATA_UNTIL,
            VariableIdentifiers.CONTAINS_DATA_UNTIL,
            "2008-01-01",
        ),
    ],
)
def test_variables_values_can_be_changed_after_inherit_dataset_value(
    dataset_value,
    dataset_identifier,
    variable_identifier,
    update_value,
    metadata: Datadoc,
):
    state.metadata = metadata
    setattr(
        state.metadata.dataset,
        dataset_identifier,
        dataset_value,
    )
    set_variables_values_inherit_dataset_values(
        dataset_value,
        dataset_identifier,
    )
    for val in state.metadata.variables:
        assert dataset_value == get_metadata_and_stringify(
            metadata.variables_lookup[val.short_name],
            variable_identifier,
        )
    setattr(
        state.metadata.variables_lookup["pers_id"],
        variable_identifier,
        update_value,
    )
    assert dataset_value == get_metadata_and_stringify(
        metadata.variables_lookup["sivilstand"],
        variable_identifier.value,
    )
    assert dataset_value != get_metadata_and_stringify(
        metadata.variables_lookup["pers_id"],
        variable_identifier.value,
    )


def test_variables_values_multilanguage_inherit_dataset_values(
    metadata: Datadoc,
):
    state.metadata = metadata
    dataset_population_description = "Personer bosatt i Norge"
    dataset_population_description_language_item = [
        model.LanguageStringTypeItem(
            languageCode="nb",
            languageText="Personer bosatt i Norge",
        ),
    ]
    metadata_identifier = DatasetIdentifiers.POPULATION_DESCRIPTION
    language = "nb"
    setattr(
        state.metadata.dataset,
        metadata_identifier,
        dataset_population_description_language_item,
    )
    set_variables_value_multilanguage_inherit_dataset_values(
        dataset_population_description,
        metadata_identifier,
        language,
    )
    for val in state.metadata.variables:
        assert metadata.dataset.population_description == get_standard_metadata(
            metadata.variables_lookup[val.short_name],
            VariableIdentifiers.POPULATION_DESCRIPTION.value,
        )


def test_variables_values_multilanguage_can_be_changed_after_inherit_dataset_value(
    metadata: Datadoc,
):
    state.metadata = metadata
    dataset_population_description = "Persons in Norway"
    dataset_population_description_language_item = [
        model.LanguageStringTypeItem(
            languageCode="en",
            languageText="Persons in Norway",
        ),
    ]
    dataset_identifier = DatasetIdentifiers.POPULATION_DESCRIPTION
    variables_identifier = VariableIdentifiers.POPULATION_DESCRIPTION
    language = "en"
    setattr(
        state.metadata.dataset,
        dataset_identifier,
        dataset_population_description_language_item,
    )
    set_variables_value_multilanguage_inherit_dataset_values(
        dataset_population_description,
        dataset_identifier,
        language,
    )
    for val in state.metadata.variables:
        assert metadata.dataset.population_description == get_standard_metadata(
            metadata.variables_lookup[val.short_name],
            variables_identifier,
        )
    variables_language_item = [
        model.LanguageStringTypeItem(
            languageCode="en",
            languageText="Persons in Sweden",
        ),
    ]
    setattr(
        state.metadata.variables_lookup["pers_id"],
        variables_identifier,
        variables_language_item,
    )
    assert metadata.dataset.population_description != get_standard_metadata(
        metadata.variables_lookup["pers_id"],
        variables_identifier,
    )
    assert metadata.dataset.population_description == get_standard_metadata(
        metadata.variables_lookup["sivilstand"],
        variables_identifier,
    )


def test_variables_values_inherit_dataset_date_values_derived_from_path(
    metadata: Datadoc,
):
    state.metadata = metadata
    dataset_contains_data_from = "2021-10-10"
    setattr(
        state.metadata.dataset,
        DatasetIdentifiers.CONTAINS_DATA_FROM,
        dataset_contains_data_from,
    )
    set_variables_values_inherit_dataset_derived_date_values()
    for val in state.metadata.variables:
        assert metadata.dataset.contains_data_from == get_standard_metadata(
            metadata.variables_lookup[val.short_name],
            VariableIdentifiers.CONTAINS_DATA_FROM,
        )
    for val in state.metadata.variables:
        assert metadata.variables_lookup[val.short_name].contains_data_until is None
    setattr(
        state.metadata.variables_lookup["pers_id"],
        VariableIdentifiers.CONTAINS_DATA_FROM,
        "2011-10-10",
    )
    set_variables_values_inherit_dataset_derived_date_values()
    assert metadata.variables_lookup[
        "pers_id"
    ].contains_data_from != get_standard_metadata(
        metadata.dataset,
        DatasetIdentifiers.CONTAINS_DATA_FROM,
    )


def test_variables_values_inherit_dataset_date_values_not_when_variable_has_value(
    metadata: Datadoc,
):
    state.metadata = metadata
    dataset_contains_data_until = "2024-01-01"
    setattr(
        state.metadata.variables_lookup["pers_id"],
        VariableIdentifiers.CONTAINS_DATA_UNTIL,
        "2011-12-10",
    )
    setattr(
        state.metadata.dataset,
        DatasetIdentifiers.CONTAINS_DATA_UNTIL,
        dataset_contains_data_until,
    )
    set_variables_values_inherit_dataset_derived_date_values()
    assert metadata.variables_lookup[
        "pers_id"
    ].contains_data_until != get_standard_metadata(
        metadata.dataset,
        DatasetIdentifiers.CONTAINS_DATA_UNTIL,
    )


def test_variables_metadata_control_return_alert(metadata: Datadoc):
    """Return alert when obligatory metadata is missing."""
    state.metadata = metadata
    missing_metadata: list = []
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        state.metadata.write_metadata_document()
        if issubclass(w[1].category, ObligatoryVariableWarning):
            missing_metadata.append(str(w[1].message))
    result = variables_control(missing_metadata, metadata.variables)
    assert isinstance(result, dbc.Alert)


def test_variables_metadata_control_dont_return_alert(metadata: Datadoc):
    state.metadata = metadata
    missing_metadata: list[str] = []
    for val in state.metadata.variables:
        """Not return alert when all obligatory metadata has value."""
        setattr(
            state.metadata.variables_lookup[val.short_name],
            VariableIdentifiers.NAME,
            model.LanguageStringType(
                [model.LanguageStringTypeItem(languageCode="nb", languageText="Test")],
            ),
        )
        setattr(
            state.metadata.variables_lookup[val.short_name],
            VariableIdentifiers.DATA_TYPE,
            enums.DataType.STRING,
        )
        setattr(
            state.metadata.variables_lookup[val.short_name],
            VariableIdentifiers.VARIABLE_ROLE,
            enums.VariableRole.MEASURE,
        )
        setattr(
            state.metadata.variables_lookup[val.short_name],
            VariableIdentifiers.DEFINITION_URI,
            "https://www.hat.com",
        )
        setattr(
            state.metadata.variables_lookup[val.short_name],
            VariableIdentifiers.IS_PERSONAL_DATA,
            True,
        )
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        state.metadata.write_metadata_document()
        if issubclass(w[0].category, ObligatoryVariableWarning):
            missing_metadata.append(str(w[0].message))
    result = variables_control(missing_metadata, metadata.variables)
    assert result is None


def test_accept_variable_metadata_input_when_shortname_is_non_ascii(
    metadata_illegal_shortnames: Datadoc,
):
    state.metadata = metadata_illegal_shortnames
    assert metadata_illegal_shortnames.variables[-1].short_name == "rådyr"
    assert (
        accept_variable_metadata_input(
            "Format value",
            metadata_illegal_shortnames.variables[-1].short_name,
            metadata_field=VariableIdentifiers.FORMAT.value,
            language="nb",
        )
        is None
    )

    assert (
        getattr(state.metadata.variables[-1], VariableIdentifiers.FORMAT.value)
        == "Format value"
    )


@pytest.mark.parametrize(
    ("metadata_field", "value", "expected_model_value"),
    [
        (
            PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME,
            "2024-12-31",
            datetime.datetime(
                2024,
                12,
                31,
                0,
                0,
                0,
                tzinfo=datetime.UTC,
            ),
        ),
        (
            PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME,
            "",
            None,
        ),
        (
            PseudoVariableIdentifiers.PSEUDONYMIZATION_TIME,
            None,
            None,
        ),
        (
            PseudoVariableIdentifiers.STABLE_IDENTIFIER_VERSION,
            "stable identifier ",
            "stable identifier",
        ),
        (
            PseudoVariableIdentifiers.STABLE_IDENTIFIER_TYPE,
            "",
            None,
        ),
        (
            PseudoVariableIdentifiers.ENCRYPTION_ALGORITHM,
            "TINK-FPE",
            "TINK-FPE",
        ),
        (
            PseudoVariableIdentifiers.ENCRYPTION_KEY_REFERENCE,
            "SSB_GLOBAL_KEY_1",
            "SSB_GLOBAL_KEY_1",
        ),
    ],
)
def test_accept_pseudo_variable_metadata_input_valid(
    metadata: Datadoc,
    metadata_field: PseudoVariableIdentifiers,
    value: str | datetime.datetime | None,
    expected_model_value: Any,  # noqa: ANN401
):
    state.metadata = metadata
    first_var_short_name = metadata.variables[0].short_name
    metadata.add_pseudonymization(first_var_short_name)
    assert (
        accept_pseudo_variable_metadata_input(
            value,
            first_var_short_name,
            metadata_field=metadata_field.value,
        )
        is None
    )
    variable = state.metadata.variables_lookup.get(first_var_short_name)
    assert variable is not None
    assert (
        getattr(
            variable.pseudonymization,
            metadata_field.value,
        )
        == expected_model_value
    )


@dataclass
class PseudoCase:
    """Test cases Pseudonymization."""

    selected_algorithm: enums.PseudonymizationAlgorithmsEnum | None | str
    expected_workspace_type: dbc.Form | list
    expected_number_editable_inputs: int
    expected_identifiers_in_workspace: list | None
    expected_variable_pseudonymization: bool
    saved_pseudonymization: model.Pseudonymization | None = None
    expected_algorithm_parameters_length: int | None = None


@pytest.mark.parametrize(
    "case",
    [
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=1,
            expected_identifiers_in_workspace=["pseudonymization_time"],
            expected_variable_pseudonymization=True,
            expected_algorithm_parameters_length=2,
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=2,
            expected_identifiers_in_workspace=[
                "pseudonymization_time",
                "stable_identifier_version",
            ],
            expected_variable_pseudonymization=True,
            expected_algorithm_parameters_length=3,
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=1,
            expected_identifiers_in_workspace=["pseudonymization_time"],
            expected_variable_pseudonymization=True,
            expected_algorithm_parameters_length=1,
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.CUSTOM,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=5,
            expected_identifiers_in_workspace=["pseudonymization_time"],
            expected_variable_pseudonymization=True,
            expected_algorithm_parameters_length=None,
        ),
        PseudoCase(
            selected_algorithm=None,
            expected_workspace_type=list,
            expected_number_editable_inputs=0,
            expected_identifiers_in_workspace=None,
            expected_variable_pseudonymization=False,
        ),
        PseudoCase(
            selected_algorithm=None,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=1,
            expected_identifiers_in_workspace=["pseudonymization_time"],
            expected_variable_pseudonymization=True,
            saved_pseudonymization=model.Pseudonymization(
                encryption_algorithm=constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION
            ),
        ),
    ],
    ids=[
        "PAPIS without stable ID",
        "PAPIS without stable ID",
        "DAEAD",
        "Custom",
        "No algorithm selected",
        "None",
    ],
)
def test_populate_pseudonymization_workspace(
    case,
    metadata: Datadoc,
):
    state.metadata = metadata
    first_var_short_name = metadata.variables[0].short_name
    variable = state.metadata.variables_lookup.get(first_var_short_name)
    assert variable is not None
    if case.saved_pseudonymization:
        variable.pseudonymization = case.saved_pseudonymization
    pseudonymization_workspace = populate_pseudo_workspace(
        variable, case.selected_algorithm
    )
    assert pseudonymization_workspace is not None
    assert isinstance(pseudonymization_workspace, case.expected_workspace_type)
    if case.expected_number_editable_inputs > 0:
        assert (
            len(pseudonymization_workspace.children)
            == case.expected_number_editable_inputs
        )
        all_ids = [child.id["id"] for child in pseudonymization_workspace.children]
        for ids in case.expected_identifiers_in_workspace:
            assert ids in all_ids
    if case.expected_variable_pseudonymization is True:
        assert variable.pseudonymization is not None
        if variable.pseudonymization.encryption_algorithm_parameters is not None:
            assert case.expected_algorithm_parameters_length == len(
                variable.pseudonymization.encryption_algorithm_parameters
            )
    else:
        assert variable.pseudonymization is None


@pytest.mark.parametrize(
    "case",
    [
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=1,
            expected_identifiers_in_workspace=["pseudonymization_time"],
            expected_variable_pseudonymization=True,
            saved_pseudonymization=model.Pseudonymization(
                encryption_algorithm=constants.PAPIS_ALGORITHM_ENCRYPTION,
                encryption_key_reference=constants.PAPIS_ENCRYPTION_KEY_REFERENCE,
                encryption_algorithm_parameters=[
                    {
                        constants.ENCRYPTION_PARAMETER_KEY_ID: constants.PAPIS_ENCRYPTION_KEY_REFERENCE
                    },
                    {
                        constants.ENCRYPTION_PARAMETER_STRATEGY: constants.ENCRYPTION_PARAMETER_STRATEGY_SKIP
                    },
                ],
            ),
            expected_algorithm_parameters_length=1,
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=2,
            expected_identifiers_in_workspace=["pseudonymization_time"],
            expected_variable_pseudonymization=True,
            saved_pseudonymization=model.Pseudonymization(
                encryption_algorithm=constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION,
                encryption_key_reference=constants.DAEAD_ENCRYPTION_KEY_REFERENCE,
                encryption_algorithm_parameters=[
                    {
                        constants.ENCRYPTION_PARAMETER_KEY_ID: constants.DAEAD_ENCRYPTION_KEY_REFERENCE
                    }
                ],
            ),
            expected_algorithm_parameters_length=3,
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.CUSTOM,
            expected_workspace_type=dbc.Form,
            expected_number_editable_inputs=5,
            expected_identifiers_in_workspace=["pseudonymization_time"],
            expected_variable_pseudonymization=True,
            saved_pseudonymization=model.Pseudonymization(
                encryption_algorithm=constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION,
                pseudonymization_time=datetime.datetime(
                    2024, 12, 31, 0, 0, 0, tzinfo=datetime.UTC
                ),
                encryption_algorithm_parameters=[
                    {
                        constants.ENCRYPTION_PARAMETER_KEY_ID: constants.DAEAD_ENCRYPTION_KEY_REFERENCE
                    }
                ],
            ),
            expected_algorithm_parameters_length=0,
        ),
    ],
    ids=[
        "Change from PAPIS without stable ID to DAEAD",
        "Change from DAEAD to PAPIS with stable ID",
        "Change from DAEAD to CUSTOM",
    ],
)
def test_update_pseudonymization_algorithm(case, metadata: Datadoc):
    state.metadata = metadata
    first_var_short_name = metadata.variables[0].short_name
    variable = state.metadata.variables_lookup.get(first_var_short_name)
    assert variable is not None
    if case.saved_pseudonymization:
        variable.pseudonymization = case.saved_pseudonymization
    update_selected_pseudonymization(
        variable,
        case.saved_pseudonymization.encryption_algorithm,
        case.selected_algorithm,
    )
    if case.expected_variable_pseudonymization is True:
        assert variable.pseudonymization is not None
        assert (
            variable.pseudonymization.pseudonymization_time
            == case.saved_pseudonymization.pseudonymization_time
        )
        if variable.pseudonymization.encryption_algorithm_parameters is not None:
            assert case.expected_algorithm_parameters_length == len(
                variable.pseudonymization.encryption_algorithm_parameters
            )


def test_delete_pseudonymization(
    metadata: Datadoc,
):
    state.metadata = metadata
    first_var_short_name = metadata.variables[0].short_name
    variable = state.metadata.variables_lookup.get(first_var_short_name)
    assert variable is not None
    variable.pseudonymization = model.Pseudonymization(
        encryption_algorithm=constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION
    )
    assert variable.pseudonymization.encryption_algorithm == "TINK-DAEAD"
    mutate_variable_pseudonymization(variable, constants.DELETE_SELECTED)
    assert variable.pseudonymization is None



"""Configuration for the Gunicorn server."""

from datadoc_editor.logging_configuration.logging_config import get_log_config

bind = "0.0.0.0:8050"
workers = 1
loglevel = "info"
preload = True
logconfig_dict = get_log_config()



"""Verify that we are in sync with the Model."""

from dapla_metadata.datasets import model

from datadoc_editor.frontend.fields.display_dataset import DISPLAY_DATASET
from datadoc_editor.frontend.fields.display_dataset import DatasetIdentifiers
from datadoc_editor.frontend.fields.display_variables import DISPLAY_VARIABLES
from datadoc_editor.frontend.fields.display_variables import VariableIdentifiers


def test_dataset_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    datadoc_values = sorted([i.value for i in DatasetIdentifiers])
    model_values = sorted(model.Dataset().model_dump().keys())

    # TODO @Jorgen-5: Fields that are currently not supported by datadoc # noqa: TD003
    model_values.remove("custom_type")

    assert datadoc_values == model_values
    assert sorted(DatasetIdentifiers) == sorted(DISPLAY_DATASET.keys())


def test_variables_metadata_definition_parity():
    """The metadata fields are currently defined in multiple places for technical reasons. We want these to always be exactly identical."""
    datadoc_values = sorted([i.value for i in VariableIdentifiers])
    model_values = sorted(model.Variable().model_dump().keys())

    # TODO @Jorgen-5: Fields that are currently not supported by datadoc # noqa: TD003
    model_values.remove("custom_type")
    model_values.remove("special_value")
    model_values.remove("pseudonymization")

    assert datadoc_values == model_values

    assert sorted(VariableIdentifiers) == sorted(DISPLAY_VARIABLES.keys())
